import json
import logging
from collections.abc import Iterable
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from hashlib import sha256

from databricks.labs.lsql.backends import SqlBackend
from databricks.sdk import WorkspaceClient
from databricks.sdk.service import compute
from databricks.sdk.service.compute import ClusterDetails
from databricks.sdk.service.jobs import (
    BaseJob,
    BaseRun,
    DbtTask,
    GitSource,
    PythonWheelTask,
    RunConditionTask,
    RunTask,
    RunType,
    SparkJarTask,
    SqlTask,
)

from databricks.labs.ucx.assessment.clusters import CheckClusterMixin
from databricks.labs.ucx.assessment.crawlers import spark_version_compatibility
from databricks.labs.ucx.framework.crawlers import CrawlerBase
from databricks.labs.ucx.framework.utils import escape_sql_identifier

logger = logging.getLogger(__name__)


@dataclass
class JobInfo:
    job_id: str
    success: int
    failures: str
    job_name: str | None = None
    creator: str | None = None


class JobsMixin:
    @classmethod
    def _get_cluster_configs_from_all_jobs(cls, all_jobs, all_clusters_by_id):
        for job in all_jobs:
            if job.settings is None:
                continue
            if job.settings.job_clusters is not None:
                yield from cls._job_clusters(job)
            if job.settings.tasks is None:
                continue
            yield from cls._task_clusters(job, all_clusters_by_id)

    @classmethod
    def _task_clusters(cls, job, all_clusters_by_id):
        for task in job.settings.tasks:
            if task.existing_cluster_id is not None:
                interactive_cluster = all_clusters_by_id.get(task.existing_cluster_id, None)
                if interactive_cluster is None:
                    continue
                yield job, interactive_cluster
            elif task.new_cluster is not None:
                yield job, task.new_cluster

    @staticmethod
    def _job_clusters(job):
        for job_cluster in job.settings.job_clusters:
            if job_cluster.new_cluster is None:
                continue
            yield job, job_cluster.new_cluster


class JobsCrawler(CrawlerBase[JobInfo], JobsMixin, CheckClusterMixin):
    def __init__(self, ws: WorkspaceClient, sbe: SqlBackend, schema):
        super().__init__(sbe, "hive_metastore", schema, "jobs", JobInfo)
        self._ws = ws

    def _crawl(self) -> Iterable[JobInfo]:
        all_jobs = list(self._ws.jobs.list(expand_tasks=True))
        all_clusters = {c.cluster_id: c for c in self._ws.clusters.list()}
        return self._assess_jobs(all_jobs, all_clusters)

    def _assess_jobs(self, all_jobs: list[BaseJob], all_clusters_by_id) -> Iterable[JobInfo]:
        job_assessment, job_details = self._prepare(all_jobs)
        for job, cluster_config in self._get_cluster_configs_from_all_jobs(all_jobs, all_clusters_by_id):
            job_id = job.job_id
            if not job_id:
                continue
            cluster_details = ClusterDetails.from_dict(cluster_config.as_dict())
            cluster_failures = self._check_cluster_failures(cluster_details, "Job cluster")
            cluster_failures.extend(self._check_jar_task(job.settings.tasks))
            job_assessment[job_id].update(cluster_failures)

        # TODO: next person looking at this - rewrite, as this code makes no sense
        for job_key in job_details.keys():  # pylint: disable=consider-using-dict-items,consider-iterating-dictionary
            job_details[job_key].failures = json.dumps(list(job_assessment[job_key]))
            if len(job_assessment[job_key]) > 0:
                job_details[job_key].success = 0
        return list(job_details.values())

    @staticmethod
    def _prepare(all_jobs) -> tuple[dict[int, set[str]], dict[int, JobInfo]]:
        job_assessment: dict[int, set[str]] = {}
        job_details: dict[int, JobInfo] = {}
        for job in all_jobs:
            if not job.job_id:
                continue
            job_assessment[job.job_id] = set()
            if not job.creator_user_name:
                logger.warning(
                    f"Job {job.job_id} have Unknown creator, it means that the original creator has been deleted "
                    f"and should be re-created"
                )

            job_settings = job.settings
            if not job_settings:
                continue
            job_name = job_settings.name
            if not job_name:
                job_name = "Unknown"

            job_details[job.job_id] = JobInfo(
                job_id=str(job.job_id),
                job_name=job_name,
                creator=job.creator_user_name,
                success=1,
                failures="[]",
            )
        return job_assessment, job_details

    def snapshot(self) -> Iterable[JobInfo]:
        return self._snapshot(self._try_fetch, self._crawl)

    def _try_fetch(self) -> Iterable[JobInfo]:
        for row in self._fetch(f"SELECT * FROM {escape_sql_identifier(self.full_name)}"):
            yield JobInfo(*row)

    def _check_jar_task(self, all_task: list[RunTask]) -> list[str]:
        task_failures: list[str] = []
        for task in all_task:
            if task.spark_jar_task:
                task_failures.append(f"task {task.task_key} is a jar task")
        return task_failures


@dataclass
class SubmitRunInfo:
    run_ids: str  # JSON-encoded list of run ids
    hashed_id: str  # a pseudo id that combines all the hashable attributes of the run
    failures: str = "[]"  # JSON-encoded list of failures


class SubmitRunsCrawler(CrawlerBase[SubmitRunInfo], JobsMixin, CheckClusterMixin):
    _FS_LEVEL_CONF_SETTING_PATTERNS = [
        "fs.s3a",
        "fs.s3n",
        "fs.s3",
        "fs.azure",
        "fs.wasb",
        "fs.abfs",
        "fs.adl",
    ]

    def __init__(self, ws: WorkspaceClient, sbe: SqlBackend, schema: str, num_days_history: int):
        super().__init__(sbe, "hive_metastore", schema, "submit_runs", SubmitRunInfo)
        self._ws = ws
        self._num_days_history = num_days_history

    def snapshot(self) -> Iterable[SubmitRunInfo]:
        return self._snapshot(self._try_fetch, self._crawl)

    @staticmethod
    def _dt_to_ms(date_time: datetime):
        return int(date_time.timestamp() * 1000)

    @staticmethod
    def _get_current_dttm() -> datetime:
        return datetime.now(timezone.utc)

    def _crawl(self) -> Iterable[SubmitRunInfo]:
        end = self._dt_to_ms(self._get_current_dttm())
        start = self._dt_to_ms(self._get_current_dttm() - timedelta(days=self._num_days_history))
        submit_runs = self._ws.jobs.list_runs(
            expand_tasks=True,
            completed_only=True,
            run_type=RunType.SUBMIT_RUN,
            start_time_from=start,
            start_time_to=end,
        )
        all_clusters = {c.cluster_id: c for c in self._ws.clusters.list()}
        return self._assess_job_runs(submit_runs, all_clusters)

    def _try_fetch(self) -> Iterable[SubmitRunInfo]:
        for row in self._fetch(f"SELECT * FROM {escape_sql_identifier(self.full_name)}"):
            yield SubmitRunInfo(*row)

    def _check_spark_conf(self, conf: dict[str, str], source: str) -> list[str]:
        failures: list[str] = []
        for key in conf.keys():
            if any(pattern in key for pattern in self._FS_LEVEL_CONF_SETTING_PATTERNS):
                failures.append(f"Potentially unsupported config property: {key}")

        failures.extend(super()._check_spark_conf(conf, source))
        return failures

    def _check_cluster_failures(self, cluster: ClusterDetails, source: str) -> list[str]:
        failures: list[str] = []
        if cluster.aws_attributes and cluster.aws_attributes.instance_profile_arn:
            failures.append(f"using instance profile: {cluster.aws_attributes.instance_profile_arn}")

        failures.extend(super()._check_cluster_failures(cluster, source))
        return failures

    @staticmethod
    def _needs_compatibility_check(spec: compute.ClusterSpec) -> bool:
        """
        # we recognize a task as a potentially incompatible one if:
        # 1. cluster is not configured with data security mode
        # 2. cluster's DBR version is greater than 11.3
        """
        if not spec.data_security_mode:
            compatibility = spark_version_compatibility(spec.spark_version)
            return compatibility == "supported"
        return False

    def _get_hash_from_run(self, run: BaseRun) -> str:
        hashable_items = []
        all_tasks: list[RunTask] = run.tasks if run.tasks is not None else []
        for task in sorted(all_tasks, key=lambda x: x.task_key if x.task_key is not None else ""):
            hashable_items.extend(self._run_task_values(task))

        if run.git_source:
            hashable_items.extend(self._git_source_values(run.git_source))

        return sha256(bytes("|".join(hashable_items).encode("utf-8"))).hexdigest()

    @classmethod
    def _sql_task_values(cls, task: SqlTask) -> list[str]:
        hash_values = [
            task.file.path if task.file else None,
            task.alert.alert_id if task.alert else None,
            task.dashboard.dashboard_id if task.dashboard else None,
            task.query.query_id if task.query else None,
        ]
        return [str(value) for value in hash_values if value is not None]

    @classmethod
    def _git_source_values(cls, source: GitSource) -> list[str]:
        hash_values = [source.git_url]
        return [str(value) for value in hash_values if value is not None]

    @classmethod
    def _dbt_task_values(cls, dbt_task: DbtTask) -> list[str]:
        hash_values = [
            dbt_task.schema,
            dbt_task.catalog,
            dbt_task.warehouse_id,
            dbt_task.project_directory,
            ",".join(sorted(dbt_task.commands)),
        ]
        return [str(value) for value in hash_values if value is not None]

    @classmethod
    def _jar_task_values(cls, spark_jar_task: SparkJarTask) -> list[str]:
        hash_values = [spark_jar_task.jar_uri, spark_jar_task.main_class_name]
        return [str(value) for value in hash_values if value is not None]

    @classmethod
    def _python_wheel_task_values(cls, pw_task: PythonWheelTask) -> list[str]:
        hash_values = [pw_task.package_name, pw_task.entry_point]
        return [str(value) for value in hash_values if value is not None]

    @classmethod
    def _run_condition_task_values(cls, c_task: RunConditionTask) -> list[str]:
        hash_values = [c_task.op.value if c_task.op else None, c_task.right, c_task.left, c_task.outcome]
        return [str(value) for value in hash_values if value is not None]

    @classmethod
    def _run_task_values(cls, task: RunTask) -> list[str]:
        """
        Retrieve all hashable attributes and append to a list with None removed
        - specifically ignore parameters as these change.
        """
        hash_values = [
            task.notebook_task.notebook_path if task.notebook_task else None,
            task.spark_python_task.python_file if task.spark_python_task else None,
            (
                '|'.join(task.spark_submit_task.parameters)
                if (task.spark_submit_task and task.spark_submit_task.parameters)
                else None
            ),
            task.pipeline_task.pipeline_id if task.pipeline_task is not None else None,
            task.run_job_task.job_id if task.run_job_task else None,
        ]
        hash_lists = [
            cls._jar_task_values(task.spark_jar_task) if task.spark_jar_task else None,
            (cls._python_wheel_task_values(task.python_wheel_task) if (task.python_wheel_task) else None),
            cls._sql_task_values(task.sql_task) if task.sql_task else None,
            cls._dbt_task_values(task.dbt_task) if task.dbt_task else None,
            cls._run_condition_task_values(task.condition_task) if task.condition_task else None,
            cls._git_source_values(task.git_source) if task.git_source else None,
        ]
        # combining all the values from the lists where the list is not "None"
        hash_values_from_lists = sum([hash_list for hash_list in hash_lists if hash_list], [])
        return [str(value) for value in hash_values + hash_values_from_lists]

    def _assess_job_runs(self, submit_runs: Iterable[BaseRun], all_clusters_by_id) -> Iterable[SubmitRunInfo]:
        """
        Assessment logic:
        1. For eaxch submit run, we analyze all tasks inside this run.
        2. Per each task, we calculate a unique hash based on the _retrieve_hash_values_from_task function
        3. Then we coalesce all task hashes into a single hash for the submit run
        4. Coalesce all runs under the same hash into a single pseudo-job
        5. Return a list of pseudo-jobs with their assessment results
        """
        result: dict[str, SubmitRunInfo] = {}
        runs_per_hash: dict[str, list[int | None]] = {}

        for submit_run in submit_runs:
            task_failures: list[str] = []
            # v2.1+ API, with tasks
            if submit_run.tasks:
                self._check_run_task(submit_run.tasks, all_clusters_by_id, task_failures)
            # v2.0 API, without tasks
            elif submit_run.cluster_spec:
                _cluster_details = ClusterDetails.from_dict(submit_run.cluster_spec.as_dict())
                task_failures.extend(self._check_cluster_failures(_cluster_details, "root_task"))
            hashed_id = self._get_hash_from_run(submit_run)
            if hashed_id in runs_per_hash:
                runs_per_hash[hashed_id].append(submit_run.run_id)
            else:
                runs_per_hash[hashed_id] = [submit_run.run_id]
            result[hashed_id] = SubmitRunInfo(
                run_ids=json.dumps(runs_per_hash[hashed_id]),
                hashed_id=hashed_id,
                failures=json.dumps(list(set(task_failures))),
            )

        return list(result.values())

    def _check_run_task(self, all_tasks: list[RunTask], clusters: dict[str, ClusterDetails], task_failures: list[str]):
        for task in sorted(all_tasks, key=lambda x: x.task_key if x.task_key is not None else ""):
            _task_key = task.task_key if task.task_key is not None else ""
            cluster_details = None
            if task.new_cluster:
                cluster_details = ClusterDetails.from_dict(task.new_cluster.as_dict())
                if self._needs_compatibility_check(task.new_cluster):
                    task_failures.append("no data security mode specified")
            if task.existing_cluster_id:
                cluster_details = clusters.get(task.existing_cluster_id, None)
            if cluster_details:
                task_failures.extend(self._check_cluster_failures(cluster_details, _task_key))
            if task.spark_jar_task:
                task_failures.append(f"task {task.task_key} is a jar task")



import logging
import os
import shutil
from functools import cached_property

from databricks.labs.lsql.backends import SqlBackend, StatementExecutionBackend
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import NotFound

from databricks.labs.ucx.assessment.aws import AWSResources
from databricks.labs.ucx.framework.utils import run_command
from databricks.labs.ucx.aws.access import AWSResourcePermissions
from databricks.labs.ucx.aws.credentials import IamRoleMigration, IamRoleCreation
from databricks.labs.ucx.aws.locations import AWSExternalLocationsMigration
from databricks.labs.ucx.azure.access import AzureResourcePermissions
from databricks.labs.ucx.azure.credentials import StorageCredentialManager, ServicePrincipalMigration
from databricks.labs.ucx.azure.locations import ExternalLocationsMigration
from databricks.labs.ucx.azure.resources import AzureAPIClient, AzureResources
from databricks.labs.ucx.contexts.application import CliContext
from databricks.labs.ucx.hive_metastore.migration_status import MigrationIndex
from databricks.labs.ucx.source_code.base import CurrentSessionState
from databricks.labs.ucx.source_code.linters.context import LinterContext
from databricks.labs.ucx.source_code.linters.files import LocalFileMigrator, LocalCodeLinter
from databricks.labs.ucx.source_code.notebooks.loaders import NotebookLoader
from databricks.labs.ucx.workspace_access.clusters import ClusterAccess

logger = logging.getLogger(__name__)


class WorkspaceContext(CliContext):
    def __init__(self, ws: WorkspaceClient, named_parameters: dict[str, str] | None = None):
        super().__init__(named_parameters)
        self._ws = ws

    @cached_property
    def workspace_client(self) -> WorkspaceClient:
        return self._ws

    @cached_property
    def sql_backend(self) -> SqlBackend:
        return StatementExecutionBackend(self.workspace_client, self.config.warehouse_id)

    @cached_property
    def cluster_access(self):
        return ClusterAccess(self.installation, self.workspace_client, self.prompts)

    @cached_property
    def azure_cli_authenticated(self):
        if not self.is_azure:
            raise NotImplementedError("Azure only")
        if self.connect_config.auth_type != "azure-cli":
            raise ValueError("In order to obtain AAD token, Please run azure cli to authenticate.")
        return True

    @cached_property
    def azure_management_client(self):
        if not self.azure_cli_authenticated:
            raise NotImplementedError
        return AzureAPIClient(
            self.workspace_client.config.arm_environment.resource_manager_endpoint,
            self.workspace_client.config.arm_environment.service_management_endpoint,
        )

    @cached_property
    def microsoft_graph_client(self):
        if not self.azure_cli_authenticated:
            raise NotImplementedError
        return AzureAPIClient("https://graph.microsoft.com", "https://graph.microsoft.com")

    @cached_property
    def azure_subscription_id(self):
        subscription_id = self.named_parameters.get("subscription_id")
        if not subscription_id:
            raise ValueError("Please enter subscription id to scan storage accounts in.")
        return subscription_id

    @cached_property
    def azure_resources(self):
        return AzureResources(
            self.azure_management_client,
            self.microsoft_graph_client,
            [self.azure_subscription_id],
        )

    @cached_property
    def azure_resource_permissions(self):
        return AzureResourcePermissions(
            self.installation,
            self.workspace_client,
            self.azure_resources,
            self.external_locations,
        )

    @cached_property
    def azure_credential_manager(self):
        return StorageCredentialManager(self.workspace_client)

    @cached_property
    def service_principal_migration(self):
        return ServicePrincipalMigration(
            self.installation,
            self.workspace_client,
            self.azure_resource_permissions,
            self.azure_service_principal_crawler,
            self.azure_credential_manager,
        )

    @cached_property
    def external_locations_migration(self):
        if self.is_aws:
            return AWSExternalLocationsMigration(
                self.workspace_client,
                self.external_locations,
                self.aws_resource_permissions,
                self.principal_acl,
            )
        if self.is_azure:
            return ExternalLocationsMigration(
                self.workspace_client,
                self.external_locations,
                self.azure_resource_permissions,
                self.azure_resources,
                self.principal_acl,
            )
        raise NotImplementedError

    @cached_property
    def aws_cli_run_command(self):
        # this is a convenience method for unit testing
        if not shutil.which("aws"):
            raise ValueError("Couldn't find AWS CLI in path. Please install the CLI from https://aws.amazon.com/cli/")
        return run_command

    @cached_property
    def aws_profile(self):
        aws_profile = self.named_parameters.get("aws_profile")
        if not aws_profile:
            aws_profile = os.getenv("AWS_DEFAULT_PROFILE")
        if not aws_profile:
            raise ValueError(
                "AWS Profile is not specified. Use the environment variable [AWS_DEFAULT_PROFILE] "
                "or use the '--aws-profile=[profile-name]' parameter."
            )
        return aws_profile

    @cached_property
    def aws_resources(self):
        if not self.is_aws:
            raise NotImplementedError("AWS only")
        return AWSResources(self.aws_profile, self.aws_cli_run_command)

    @cached_property
    def aws_resource_permissions(self):
        return AWSResourcePermissions(
            self.installation,
            self.workspace_client,
            self.aws_resources,
            self.external_locations,
            self.named_parameters.get("kms_key"),
        )

    @cached_property
    def iam_role_migration(self):
        return IamRoleMigration(
            self.installation,
            self.aws_resource_permissions,
            self.iam_credential_manager,
        )

    @cached_property
    def iam_role_creation(self):
        return IamRoleCreation(
            self.installation,
            self.workspace_client,
            self.aws_resource_permissions,
        )

    @cached_property
    def notebook_loader(self) -> NotebookLoader:
        return NotebookLoader()


class LocalCheckoutContext(WorkspaceContext):
    """Local context extends Workspace context to provide extra properties
    for running local operations."""

    def linter_context_factory(self, session_state: CurrentSessionState | None = None):
        try:
            index = self.tables_migrator.index()
        except NotFound:
            logger.warning("Metastore does not seem to exist yet. Skipping loading of migration status.")
            index = MigrationIndex([])
        if session_state is None:
            session_state = CurrentSessionState()
        return LinterContext(index, session_state)

    @cached_property
    def local_file_migrator(self):
        return LocalFileMigrator(lambda: self.linter_context_factory(CurrentSessionState()))

    @cached_property
    def local_code_linter(self):
        session_state = CurrentSessionState()
        return LocalCodeLinter(
            self.file_loader,
            self.folder_loader,
            self.path_lookup,
            session_state,
            self.dependency_resolver,
            lambda: self.linter_context_factory(session_state),
        )



from collections.abc import Iterator

from databricks.labs.lsql.backends import SqlBackend
from databricks.labs.lsql.core import Row

from .base import TableIdentifier, TableMetadata, ColumnMetadata, TableMetadataRetriever


class DatabricksTableMetadataRetriever(TableMetadataRetriever):
    def __init__(self, sql_backend: SqlBackend):
        self._sql_backend = sql_backend

    def get_metadata(self, entity: TableIdentifier) -> TableMetadata:
        """
        This method retrieves the metadata for a given table. It takes a TableIdentifier object as input,
        which represents the table for which the metadata is to be retrieved.

        Note: This method does not handle exceptions raised during the execution of the SQL query. These exceptions are
        expected to be handled by the caller in a manner appropriate for their context.
        """
        schema_query = self._build_metadata_query(entity)
        query_result: Iterator[Row] = self._sql_backend.fetch(schema_query)
        # The code uses a set comprehension to automatically deduplicate the column metadata entries,
        # Partition information are typically prefixed with a # symbol,
        # so any column name starting with # is excluded from the final set of column metadata.
        # The column metadata objects are sorted by column name to ensure a consistent order.
        columns = {
            ColumnMetadata(str(row["col_name"]), str(row["data_type"]))
            for row in query_result
            if not str(row["col_name"]).startswith("#")
        }
        return TableMetadata(entity, sorted(columns, key=lambda x: x.name))

    @classmethod
    def _build_metadata_query(cls, entity: TableIdentifier) -> str:
        if entity.catalog == "hive_metastore":
            return f"DESCRIBE TABLE {entity.fqn_escaped}"

        query = f"""
            SELECT
                LOWER(column_name) AS col_name,
                full_data_type AS data_type
            FROM
                {entity.catalog_escaped}.information_schema.columns
            WHERE
                LOWER(table_catalog)='{entity.catalog}' AND
                LOWER(table_schema)='{entity.schema}' AND
                LOWER(table_name) ='{entity.table}'
            ORDER BY col_name"""

        return query



from __future__ import annotations

import locale
import logging
from collections.abc import Iterable
from functools import cached_property
from pathlib import Path
from typing import cast

from astroid import AstroidSyntaxError, Module, NodeNG  # type: ignore

from databricks.sdk.service.workspace import Language

from databricks.labs.ucx.hive_metastore.migration_status import MigrationIndex
from databricks.labs.ucx.source_code.base import (
    Advice,
    Failure,
    Linter,
    PythonSequentialLinter,
    CurrentSessionState,
    Advisory,
    guess_encoding,
    file_language,
    is_a_notebook,
)

from databricks.labs.ucx.source_code.graph import (
    SourceContainer,
    DependencyGraph,
    DependencyProblem,
    InheritedContext,
)
from databricks.labs.ucx.source_code.linters.context import LinterContext
from databricks.labs.ucx.source_code.linters.imports import (
    SysPathChange,
    UnresolvedPath,
)
from databricks.labs.ucx.source_code.linters.python_ast import Tree, NodeBase
from databricks.labs.ucx.source_code.notebooks.cells import (
    CellLanguage,
    Cell,
    CELL_SEPARATOR,
    NOTEBOOK_HEADER,
    RunCell,
    PythonCell,
    MagicLine,
    RunCommand,
)
from databricks.labs.ucx.source_code.path_lookup import PathLookup

logger = logging.getLogger(__name__)


class Notebook(SourceContainer):

    @staticmethod
    def parse(path: Path, source: str, default_language: Language) -> Notebook:
        default_cell_language = CellLanguage.of_language(default_language)
        cells = default_cell_language.extract_cells(source)
        if cells is None:
            raise ValueError(f"Could not parse Notebook: {path}")
        return Notebook(path, source, default_language, cells, source.endswith('\n'))

    def __init__(self, path: Path, source: str, language: Language, cells: list[Cell], ends_with_lf: bool):
        self._path = path
        self._source = source
        self._language = language
        self._cells = cells
        self._ends_with_lf = ends_with_lf

    @property
    def path(self) -> Path:
        return self._path

    @property
    def cells(self) -> list[Cell]:
        return self._cells

    @property
    def original_code(self) -> str:
        return self._source

    def to_migrated_code(self):
        default_language = CellLanguage.of_language(self._language)
        header = f"{default_language.comment_prefix} {NOTEBOOK_HEADER}"
        sources = [header]
        for i, cell in enumerate(self._cells):
            migrated_code = cell.migrated_code
            if cell.language is not default_language:
                migrated_code = default_language.wrap_with_magic(migrated_code, cell.language)
            sources.append(migrated_code)
            if i < len(self._cells) - 1:
                sources.append('')
                sources.append(f'{default_language.comment_prefix} {CELL_SEPARATOR}')
                sources.append('')
        if self._ends_with_lf:
            sources.append('')  # following join will append lf
        return '\n'.join(sources)

    def build_dependency_graph(self, parent: DependencyGraph) -> list[DependencyProblem]:
        """Check for any problems with dependencies of the cells in this notebook.

        Returns:
            A list of found dependency problems; position information for problems is relative to the notebook source.
        """
        problems: list[DependencyProblem] = []
        for cell in self._cells:
            cell_problems = cell.build_dependency_graph(parent)
            problems.extend(cell_problems)
        return problems

    def build_inherited_context(self, graph: DependencyGraph, child_path: Path) -> InheritedContext:
        context = InheritedContext(None, False)
        for cell in self._cells:
            child = cell.build_inherited_context(graph, child_path)
            context = context.append(child, True)
            if context.found:
                return context
        return context

    def __repr__(self):
        return f"<Notebook {self._path}>"


class NotebookLinter:
    """
    Parses a Databricks notebook and then applies available linters
    to the code cells according to the language of the cell.
    """

    @classmethod
    def from_source(
        cls,
        index: MigrationIndex,
        path_lookup: PathLookup,
        session_state: CurrentSessionState,
        source: str,
        default_language: Language,
    ) -> NotebookLinter:
        ctx = LinterContext(index)
        notebook = Notebook.parse(Path(""), source, default_language)
        assert notebook is not None
        return cls(ctx, path_lookup, session_state, notebook)

    def __init__(
        self,
        context: LinterContext,
        path_lookup: PathLookup,
        session_state: CurrentSessionState,
        notebook: Notebook,
        inherited_tree: Tree | None = None,
    ):
        self._context: LinterContext = context
        self._path_lookup = path_lookup
        self._session_state = session_state
        self._notebook: Notebook = notebook
        # reuse Python linter across related files and notebook cells
        # this is required in order to accumulate statements for improved inference
        self._python_linter: PythonSequentialLinter = cast(PythonSequentialLinter, context.linter(Language.PYTHON))
        if inherited_tree is not None:
            self._python_linter.append_tree(inherited_tree)
        self._python_trees: dict[PythonCell, Tree] = {}  # the original trees to be linted

    def lint(self) -> Iterable[Advice]:
        yield from self._load_tree_from_notebook(self._notebook, True)
        for cell in self._notebook.cells:
            if not self._context.is_supported(cell.language.language):
                continue
            if isinstance(cell, PythonCell):
                tree = self._python_trees[cell]
                advices = self._python_linter.lint_tree(tree)
            else:
                linter = self._linter(cell.language.language)
                advices = linter.lint(cell.original_code)
            for advice in advices:
                yield advice.replace(
                    start_line=advice.start_line + cell.original_offset,
                    end_line=advice.end_line + cell.original_offset,
                )

    def _load_tree_from_notebook(self, notebook: Notebook, register_trees: bool) -> Iterable[Advice]:
        for cell in notebook.cells:
            if isinstance(cell, RunCell):
                yield from self._load_tree_from_run_cell(cell)
                continue
            if isinstance(cell, PythonCell):
                yield from self._load_tree_from_python_cell(cell, register_trees)
                continue

    def _load_tree_from_python_cell(self, python_cell: PythonCell, register_trees: bool) -> Iterable[Advice]:
        try:
            tree = Tree.normalize_and_parse(python_cell.original_code)
            if register_trees:
                self._python_trees[python_cell] = tree
            yield from self._load_children_from_tree(tree)
        except AstroidSyntaxError as e:
            yield Failure('syntax-error', str(e), 0, 0, 0, 0)

    def _load_children_from_tree(self, tree: Tree) -> Iterable[Advice]:
        assert isinstance(tree.node, Module)
        # look for child notebooks (and sys.path changes that might affect their loading)
        base_nodes: list[NodeBase] = []
        base_nodes.extend(self._list_run_magic_lines(tree))
        base_nodes.extend(SysPathChange.extract_from_tree(self._session_state, tree))
        if len(base_nodes) == 0:
            self._python_linter.append_tree(tree)
            return
        # append globals
        globs = cast(Module, tree.node).globals
        self._python_linter.append_globals(globs)
        # need to execute things in intertwined sequence so concat and sort them
        nodes = list(cast(Module, tree.node).body)
        base_nodes = sorted(base_nodes, key=lambda node: (node.node.lineno, node.node.col_offset))
        yield from self._load_children_with_base_nodes(nodes, base_nodes)
        # append remaining nodes
        self._python_linter.append_nodes(nodes)

    @staticmethod
    def _list_run_magic_lines(tree: Tree) -> Iterable[MagicLine]:

        def _ignore_problem(_code: str, _message: str, _node: NodeNG) -> None:
            return None

        commands, _ = MagicLine.extract_from_tree(tree, _ignore_problem)
        for command in commands:
            if isinstance(command.as_magic(), RunCommand):
                yield command

    def _load_children_with_base_nodes(self, nodes: list[NodeNG], base_nodes: list[NodeBase]):
        for base_node in base_nodes:
            yield from self._load_children_with_base_node(nodes, base_node)

    def _load_children_with_base_node(self, nodes: list[NodeNG], base_node: NodeBase):
        while len(nodes) > 0:
            node = nodes.pop(0)
            self._python_linter.append_nodes([node])
            if node.lineno < base_node.node.lineno:
                continue
            yield from self._load_children_from_base_node(base_node)

    def _load_children_from_base_node(self, base_node: NodeBase):
        if isinstance(base_node, SysPathChange):
            yield from self._mutate_path_lookup(base_node)
            return
        if isinstance(base_node, MagicLine):
            magic = base_node.as_magic()
            assert isinstance(magic, RunCommand)
            notebook = self._load_source_from_path(magic.notebook_path)
            if notebook is None:
                yield Advisory.from_node(
                    code='dependency-not-found',
                    message=f"Can't locate dependency: {magic.notebook_path}",
                    node=base_node.node,
                )
                return
            yield from self._load_tree_from_notebook(notebook, False)
            return

    def _mutate_path_lookup(self, change: SysPathChange):
        if isinstance(change, UnresolvedPath):
            yield Advisory.from_node(
                code='sys-path-cannot-compute-value',
                message=f"Can't update sys.path from {change.node.as_string()} because the expression cannot be computed",
                node=change.node,
            )
            return
        change.apply_to(self._path_lookup)

    def _load_tree_from_run_cell(self, run_cell: RunCell) -> Iterable[Advice]:
        path = run_cell.maybe_notebook_path()
        if path is None:
            return  # malformed run cell already reported
        notebook = self._load_source_from_path(path)
        if notebook is not None:
            yield from self._load_tree_from_notebook(notebook, False)

    def _load_source_from_path(self, path: Path | None):
        if path is None:
            return None  # already reported during dependency building
        resolved = self._path_lookup.resolve(path)
        if resolved is None:
            return None  # already reported during dependency building
        # TODO deal with workspace notebooks
        language = file_language(resolved)
        # we only support Python notebooks for now
        if language is not Language.PYTHON:
            logger.warning(f"Unsupported notebook language: {language}")
            return None
        source = resolved.read_text(guess_encoding(resolved))
        return Notebook.parse(path, source, language)

    def _linter(self, language: Language) -> Linter:
        if language is Language.PYTHON:
            return self._python_linter
        return self._context.linter(language)

    def _load_source_from_run_cell(self, run_cell: RunCell):
        path = run_cell.maybe_notebook_path()
        if path is None:
            return  # malformed run cell already reported
        resolved = self._path_lookup.resolve(path)
        if resolved is None:
            return  # already reported during dependency building
        language = file_language(resolved)
        if language is not Language.PYTHON:
            return
        source = resolved.read_text(guess_encoding(resolved))
        notebook = Notebook.parse(path, source, language)
        for cell in notebook.cells:
            if isinstance(cell, RunCell):
                self._load_source_from_run_cell(cell)
                continue
            if not isinstance(cell, PythonCell):
                continue
            self._python_linter.process_child_cell(cell.original_code)

    @staticmethod
    def name() -> str:
        return "notebook-linter"


class FileLinter:
    _NOT_YET_SUPPORTED_SUFFIXES = {
        '.scala',
        '.sh',
        '.r',
    }
    _IGNORED_SUFFIXES = {
        '.json',
        '.md',
        '.txt',
        '.xml',
        '.yml',
        '.toml',
        '.cfg',
        '.bmp',
        '.gif',
        '.png',
        '.tif',
        '.tiff',
        '.svg',
        '.jpg',
        '.jpeg',
        '.pyc',
        '.whl',
        '.egg',
        '.class',
        '.iml',
        '.gz',
    }
    _IGNORED_NAMES = {
        '.ds_store',
        '.gitignore',
        '.coverage',
        'license',
        'codeowners',
        'makefile',
        'pkg-info',
        'metadata',
        'wheel',
        'record',
        'notice',
        'zip-safe',
    }

    def __init__(
        self,
        ctx: LinterContext,
        path_lookup: PathLookup,
        session_state: CurrentSessionState,
        path: Path,
        inherited_tree: Tree | None = None,
        content: str | None = None,
    ):
        self._ctx: LinterContext = ctx
        self._path_lookup = path_lookup
        self._session_state = session_state
        self._path = path
        self._inherited_tree = inherited_tree
        self._content = content

    @cached_property
    def _source_code(self) -> str:
        if self._content is None:
            self._content = self._path.read_text(guess_encoding(self._path))
        return self._content

    def lint(self) -> Iterable[Advice]:
        encoding = locale.getpreferredencoding(False)
        try:
            is_notebook = self._is_notebook()
        except FileNotFoundError:
            failure_message = f"File not found: {self._path}"
            yield Failure("file-not-found", failure_message, 0, 0, 1, 1)
            return
        except UnicodeDecodeError:
            failure_message = f"File without {encoding} encoding is not supported {self._path}"
            yield Failure("unsupported-file-encoding", failure_message, 0, 0, 1, 1)
            return
        except PermissionError:
            failure_message = f"Missing read permission for {self._path}"
            yield Failure("file-permission", failure_message, 0, 0, 1, 1)
            return

        if is_notebook:
            yield from self._lint_notebook()
        else:
            yield from self._lint_file()

    def _is_notebook(self):
        # pre-check to avoid loading unsupported content
        language = file_language(self._path)
        if not language:
            return False
        return is_a_notebook(self._path, self._source_code)

    def _lint_file(self):
        language = file_language(self._path)
        if not language:
            suffix = self._path.suffix.lower()
            if suffix in self._IGNORED_SUFFIXES or self._path.name.lower() in self._IGNORED_NAMES:
                yield from []
            elif suffix in self._NOT_YET_SUPPORTED_SUFFIXES:
                yield Failure("unsupported-language", f"Language not supported yet for {self._path}", 0, 0, 1, 1)
            else:
                yield Failure("unknown-language", f"Cannot detect language for {self._path}", 0, 0, 1, 1)
        else:
            try:
                linter = self._ctx.linter(language)
                if self._inherited_tree is not None and isinstance(linter, PythonSequentialLinter):
                    linter.append_tree(self._inherited_tree)
                yield from linter.lint(self._source_code)
            except ValueError as err:
                failure_message = f"Error while parsing content of {self._path.as_posix()}: {err}"
                yield Failure("unsupported-content", failure_message, 0, 0, 1, 1)

    def _lint_notebook(self):
        notebook = Notebook.parse(self._path, self._source_code, file_language(self._path))
        notebook_linter = NotebookLinter(
            self._ctx, self._path_lookup, self._session_state, notebook, self._inherited_tree
        )
        yield from notebook_linter.lint()



import logging

from databricks.sdk import WorkspaceClient

from databricks.labs.ucx.__about__ import __version__
from databricks.labs.ucx.install import WorkspaceInstallation

logger = logging.getLogger("databricks.labs.ucx.install")

if __name__ == "__main__":
    logger.setLevel("INFO")
    ws = WorkspaceClient(product="ucx", product_version=__version__)
    installer = WorkspaceInstallation.current(ws)
    installer.uninstall()



from databricks.labs.ucx.hive_metastore.locations import (
    ExternalLocations,
    Mounts,
    TablesInMounts,
)
from databricks.labs.ucx.hive_metastore.tables import TablesCrawler

__all__ = ["TablesCrawler", "Mounts", "ExternalLocations", "TablesInMounts"]



from abc import abstractmethod
from collections.abc import Iterator
from dataclasses import dataclass

from astroid import Attribute, Call, Const, Name, NodeNG  # type: ignore
from databricks.labs.ucx.source_code.base import (
    Advice,
    Failure,
    PythonLinter,
    CurrentSessionState,
)
from databricks.sdk.service.compute import DataSecurityMode

from databricks.labs.ucx.source_code.linters.python_ast import Tree, TreeHelper


@dataclass
class SharedClusterMatcher:
    session_state: CurrentSessionState

    def _cluster_type_str(self) -> str:
        return (
            'Unity Catalog clusters in Shared access mode'
            if not self.session_state.is_serverless
            else 'Serverless Compute'
        )

    @abstractmethod
    def lint(self, node: NodeNG) -> Iterator[Advice]:
        pass

    def lint_tree(self, root: NodeNG) -> Iterator[Advice]:
        reported_locations = set()
        for node in Tree(root).walk():
            for advice in self.lint(node):
                loc = (advice.start_line, advice.start_col)
                if loc not in reported_locations:
                    reported_locations.add(loc)
                    yield advice


class JvmAccessMatcher(SharedClusterMatcher):
    _FIELDS = [
        "_jvm",
        "_jcol",
        "_jdf",
        "_jspark",
        "_jsparkSession",
    ]

    def lint(self, node: NodeNG) -> Iterator[Advice]:
        if not isinstance(node, Attribute):
            return
        if node.attrname not in JvmAccessMatcher._FIELDS:
            return
        yield Failure.from_node(
            code='jvm-access-in-shared-clusters',
            message=f'Cannot access Spark Driver JVM on {self._cluster_type_str()}',
            node=node,
        )


class RDDApiMatcher(SharedClusterMatcher):
    _SC_METHODS = [
        # RDD creation
        "emptyRDD",
        "parallelize",
        "range",
        "makeRDD",
        # file read methods
        "binaryFiles",
        "binaryRecords",
        "hadoopFile",
        "hadoopRDD",
        "newAPIHadoopFile",
        "newAPIHadoopRDD",
        "objectFile",
        "sequenceFile",
        "textFile",
        "wholeTextFiles",
    ]

    def lint(self, node: NodeNG) -> Iterator[Advice]:
        yield from self._lint_sc(node)
        yield from self._lint_rdd_use(node)

    def _lint_rdd_use(self, node: NodeNG) -> Iterator[Advice]:
        if isinstance(node, Attribute):
            if node.attrname == 'rdd':
                yield self._rdd_failure(node)
                return
        if isinstance(node, Call) and isinstance(node.func, Attribute) and node.func.attrname == 'mapPartitions':
            yield Failure.from_node(
                code='rdd-in-shared-clusters',
                message=f'RDD APIs are not supported on {self._cluster_type_str()}. '
                f'Use mapInArrow() or Pandas UDFs instead',
                node=node,
            )

    def _lint_sc(self, node: NodeNG) -> Iterator[Advice]:
        if not isinstance(node, Call) or not isinstance(node.func, Attribute):
            return
        if node.func.attrname not in self._SC_METHODS:
            return
        function_name = TreeHelper.get_full_function_name(node)
        if not function_name or not function_name.endswith(f"sc.{node.func.attrname}"):
            return
        yield self._rdd_failure(node)

    def _rdd_failure(self, node: NodeNG) -> Advice:
        return Failure.from_node(
            code='rdd-in-shared-clusters',
            message=f'RDD APIs are not supported on {self._cluster_type_str()}. Rewrite it using DataFrame API',
            node=node,
        )


class SparkSqlContextMatcher(SharedClusterMatcher):
    _ATTRIBUTES = ["sc", "sqlContext", "sparkContext"]
    _KNOWN_REPLACEMENTS = {"getConf": "conf", "_conf": "conf"}

    def lint(self, node: NodeNG) -> Iterator[Advice]:
        if not isinstance(node, Attribute):
            return

        if isinstance(node.expr, Name) and node.expr.name in SparkSqlContextMatcher._ATTRIBUTES:
            yield self._get_advice(node, node.expr.name)
        # sparkContext can be an attribute as in df.sparkContext.getConf()
        if isinstance(node.expr, Attribute) and node.expr.attrname == 'sparkContext':
            yield self._get_advice(node, node.expr.attrname)

    def _get_advice(self, node: Attribute, name: str) -> Advice:
        if node.attrname in self._KNOWN_REPLACEMENTS:
            replacement = self._KNOWN_REPLACEMENTS[node.attrname]
            return Failure.from_node(
                code='legacy-context-in-shared-clusters',
                message=f'{name} and {node.attrname} are not supported on {self._cluster_type_str()}. '
                f'Rewrite it using spark.{replacement}',
                node=node,
            )
        return Failure.from_node(
            code='legacy-context-in-shared-clusters',
            message=f'{name} is not supported on {self._cluster_type_str()}. Rewrite it using spark',
            node=node,
        )


class LoggingMatcher(SharedClusterMatcher):
    def lint(self, node: NodeNG) -> Iterator[Advice]:
        yield from self._match_sc_set_log_level(node)
        yield from self._match_jvm_log(node)

    def _match_sc_set_log_level(self, node: NodeNG) -> Iterator[Advice]:
        if not isinstance(node, Call) or not isinstance(node.func, Attribute):
            return
        if node.func.attrname != 'setLogLevel':
            return
        function_name = TreeHelper.get_full_function_name(node)
        if not function_name or not function_name.endswith('sc.setLogLevel'):
            return

        yield Failure.from_node(
            code='spark-logging-in-shared-clusters',
            message=f'Cannot set Spark log level directly from code on {self._cluster_type_str()}. '
            f'Remove the call and set the cluster spark conf \'spark.log.level\' instead',
            node=node,
        )

    def _match_jvm_log(self, node: NodeNG) -> Iterator[Advice]:
        if not isinstance(node, Attribute):
            return
        attribute_name = TreeHelper.get_full_attribute_name(node)
        if attribute_name and attribute_name.endswith('org.apache.log4j'):
            yield Failure.from_node(
                code='spark-logging-in-shared-clusters',
                message=f'Cannot access Spark Driver JVM logger on {self._cluster_type_str()}. '
                f'Use logging.getLogger() instead',
                node=node,
            )


class UDFMatcher(SharedClusterMatcher):
    _DBR_14_2_BELOW_NOT_SUPPORTED = ["applyInPandas", "mapInPandas", "applyInPandasWithState", "udtf", "pandas_udf"]

    def lint(self, node: NodeNG) -> Iterator[Advice]:
        if not isinstance(node, Call):
            return
        function_name = TreeHelper.get_function_name(node)

        if function_name == 'registerJavaFunction':
            yield Failure.from_node(
                code='python-udf-in-shared-clusters',
                message=f'Cannot register Java UDF from Python code on {self._cluster_type_str()}. '
                f'Use a %scala cell to register the Scala UDF using spark.udf.register.',
                node=node,
            )

        if (
            function_name in UDFMatcher._DBR_14_2_BELOW_NOT_SUPPORTED
            and self.session_state.dbr_version
            and self.session_state.dbr_version < (14, 3)
        ):
            yield Failure.from_node(
                code='python-udf-in-shared-clusters',
                message=f'{function_name} require DBR 14.3 LTS or above on {self._cluster_type_str()}',
                node=node,
            )

        if function_name == 'udf' and self.session_state.dbr_version and self.session_state.dbr_version < (14, 3):
            for keyword in node.keywords:
                if keyword.arg == 'useArrow' and isinstance(keyword.value, Const) and keyword.value.value:
                    yield Failure.from_node(
                        code='python-udf-in-shared-clusters',
                        message=f'Arrow UDFs require DBR 14.3 LTS or above on {self._cluster_type_str()}',
                        node=node,
                    )


class CatalogApiMatcher(SharedClusterMatcher):
    def lint(self, node: NodeNG) -> Iterator[Advice]:
        if not isinstance(node, Attribute):
            return
        if node.attrname == 'catalog' and TreeHelper.get_full_attribute_name(node).endswith('spark.catalog'):
            yield Failure.from_node(
                code='catalog-api-in-shared-clusters',
                message=f'spark.catalog functions require DBR 14.3 LTS or above on {self._cluster_type_str()}',
                node=node,
            )


class CommandContextMatcher(SharedClusterMatcher):
    def lint(self, node: NodeNG) -> Iterator[Advice]:
        if not isinstance(node, Call):
            return
        function_name = TreeHelper.get_full_function_name(node)
        if function_name and function_name.endswith('getContext.toJson'):
            yield Failure.from_node(
                code='to-json-in-shared-clusters',
                message=f'toJson() is not available on {self._cluster_type_str()}. '
                f'Use toSafeJson() on DBR 13.3 LTS or above to get a subset of command context information.',
                node=node,
            )


class SparkConnectPyLinter(PythonLinter):
    def __init__(self, session_state: CurrentSessionState):
        if session_state.data_security_mode != DataSecurityMode.USER_ISOLATION:
            self._matchers = []
            return

        self._matchers = [
            JvmAccessMatcher(session_state=session_state),
            RDDApiMatcher(session_state=session_state),
            SparkSqlContextMatcher(session_state=session_state),
            LoggingMatcher(session_state=session_state),
            UDFMatcher(session_state=session_state),
            CommandContextMatcher(session_state=session_state),
        ]
        if session_state.dbr_version and session_state.dbr_version < (14, 3):
            self._matchers.append(CatalogApiMatcher(session_state=session_state))

    def lint_tree(self, tree: Tree) -> Iterator[Advice]:
        for matcher in self._matchers:
            yield from matcher.lint_tree(tree.node)



import datetime
import json
import logging
from collections.abc import Callable, Iterable, Iterator
from dataclasses import dataclass
from datetime import timedelta
from functools import partial

from databricks.labs.blueprint.limiter import rate_limited
from databricks.labs.blueprint.parallel import ManyError, Threads
from databricks.labs.lsql.backends import SqlBackend
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import (
    DeadlineExceeded,
    InternalError,
    InvalidParameterValue,
    NotFound,
    PermissionDenied,
    ResourceConflict,
    TemporarilyUnavailable,
)
from databricks.sdk.retries import retried
from databricks.sdk.service import iam, ml
from databricks.sdk.service.iam import PermissionLevel

from databricks.labs.ucx.framework.crawlers import CrawlerBase
from databricks.labs.ucx.framework.utils import escape_sql_identifier
from databricks.labs.ucx.workspace_access.base import AclSupport, Permissions, StaticListing
from databricks.labs.ucx.workspace_access.groups import MigrationState

logger = logging.getLogger(__name__)


@dataclass
class GenericPermissionsInfo:
    object_id: str
    request_type: str


@dataclass
class WorkspaceObjectInfo:
    path: str
    object_type: str | None = None
    object_id: str | None = None
    language: str | None = None


class Listing:
    def __init__(self, func: Callable[..., Iterable], id_attribute: str, object_type: str):
        self._func = func
        self._id_attribute = id_attribute
        self._object_type = object_type

    def object_types(self) -> set[str]:
        return {self._object_type}

    def __iter__(self):
        started = datetime.datetime.now()
        try:
            for item in self._func():
                yield GenericPermissionsInfo(getattr(item, self._id_attribute), self._object_type)
        except NotFound as e:
            logger.warning(f"Listing {self._object_type} failed: {e}")
        since = datetime.datetime.now() - started
        logger.info(f"Listed {self._object_type} in {since}")

    def __repr__(self):
        return f"Listing({self._object_type} via {self._func.__qualname__})"


class GenericPermissionsSupport(AclSupport):
    def __init__(
        self,
        ws: WorkspaceClient,
        listings: list[Listing],
        verify_timeout: timedelta | None = timedelta(minutes=1),
        # this parameter is for testing scenarios only - [{object_type}:{object_id}]
        # it will use StaticListing class to return only object ids that has the same object type
        include_object_permissions: list[str] | None = None,
    ):
        self._ws = ws
        self._listings = listings
        self._verify_timeout = verify_timeout
        self._include_object_permissions = include_object_permissions

    def get_crawler_tasks(self):
        if self._include_object_permissions:
            for item in StaticListing(self._include_object_permissions, self.object_types()):
                yield partial(self._crawler_task, item.object_type, item.object_id)
            return
        for listing in self._listings:
            for info in listing:
                yield partial(self._crawler_task, info.request_type, info.object_id)

    def object_types(self) -> set[str]:
        all_object_types = set()
        for listing in self._listings:
            for object_type in listing.object_types():
                all_object_types.add(object_type)
        return all_object_types

    def get_apply_task(self, item: Permissions, migration_state: MigrationState):
        if not self._is_item_relevant(item, migration_state):
            return None
        object_permissions = iam.ObjectPermissions.from_dict(json.loads(item.raw))
        new_acl = self._prepare_new_acl(object_permissions, migration_state)
        return partial(self._applier_task, item.object_type, item.object_id, new_acl)

    @staticmethod
    def _is_item_relevant(item: Permissions, migration_state: MigrationState) -> bool:
        # passwords and tokens are represented on the workspace-level
        if item.object_id in {"tokens", "passwords"}:
            return True
        object_permissions = iam.ObjectPermissions.from_dict(json.loads(item.raw))
        assert object_permissions.access_control_list is not None
        for acl in object_permissions.access_control_list:
            if not acl.group_name:
                continue
            if migration_state.is_in_scope(acl.group_name):
                return True
        return False

    @staticmethod
    def _response_to_request(
        acls: list[iam.AccessControlResponse] | None = None,
    ) -> list[iam.AccessControlRequest]:
        results: list[iam.AccessControlRequest] = []
        if not acls:
            return results
        for acl in acls:
            if not acl.all_permissions:
                continue
            for permission in acl.all_permissions:
                results.append(
                    iam.AccessControlRequest(
                        acl.group_name, permission.permission_level, acl.service_principal_name, acl.user_name
                    )
                )
        return results

    @rate_limited(max_requests=100)
    def _verify(self, object_type: str, object_id: str, acl: list[iam.AccessControlRequest]):
        # in-flight check for the applied permissions
        # the api might be inconsistent, therefore we need to check that the permissions were applied
        remote_permission = self._safe_get_permissions(object_type, object_id)
        if remote_permission:
            remote_permission_as_request = self._response_to_request(remote_permission.access_control_list)
            if all(elem in remote_permission_as_request for elem in acl):
                return True
            msg = (
                f"Couldn't find permission for object type {object_type} with id {object_id}\n"
                f"acl to be applied={acl}\n"
                f"acl found in the object={remote_permission_as_request}\n"
            )
            raise NotFound(msg)
        return False

    def get_verify_task(self, item: Permissions) -> Callable[[], bool]:
        acl = iam.ObjectPermissions.from_dict(json.loads(item.raw))
        if not acl.access_control_list:
            raise ValueError(
                f"Access control list not present for object type " f"{item.object_type} and object id {item.object_id}"
            )
        permissions_as_request = self._response_to_request(acl.access_control_list)
        return partial(self._verify, item.object_type, item.object_id, permissions_as_request)

    @rate_limited(max_requests=30)
    def _applier_task(self, object_type: str, object_id: str, acl: list[iam.AccessControlRequest]):
        retryable_exceptions = [InternalError, NotFound, ResourceConflict, TemporarilyUnavailable, DeadlineExceeded]

        update_retry_on_value_error = retried(
            on=retryable_exceptions, timeout=self._verify_timeout  # type: ignore[arg-type]
        )
        update_retried_check = update_retry_on_value_error(self._safe_update_permissions)
        update_retried_check(object_type, object_id, acl)

        retry_on_value_error = retried(on=retryable_exceptions, timeout=self._verify_timeout)
        retried_check = retry_on_value_error(self._verify)
        return retried_check(object_type, object_id, acl)

    @rate_limited(max_requests=100)
    def _crawler_task(self, object_type: str, object_id: str) -> Permissions | None:
        objects_with_owner_permission = ["jobs", "pipelines"]

        permissions = self._safe_get_permissions(object_type, object_id)
        if not permissions:
            logger.warning(f"Object {object_type} {object_id} doesn't have any permissions")
            return None
        if not self._object_have_owner(permissions) and object_type in objects_with_owner_permission:
            logger.warning(
                f"Object {object_type} {object_id} doesn't have any Owner and cannot be migrated "
                f"to account level groups, consider setting a new owner or deleting this object"
            )
            return None
        return Permissions(
            object_id=object_id,
            object_type=object_type,
            raw=json.dumps(permissions.as_dict()),
        )

    def _object_have_owner(self, permissions: iam.ObjectPermissions | None):
        if not permissions:
            return False
        if not permissions.access_control_list:
            return False
        for acl in permissions.access_control_list:
            if not acl.all_permissions:
                continue
            for perm in acl.all_permissions:
                if perm.permission_level == PermissionLevel.IS_OWNER:
                    return True
        return False

    def _load_as_request(self, object_type: str, object_id: str) -> list[iam.AccessControlRequest]:
        loaded = self._safe_get_permissions(object_type, object_id)
        if loaded is None:
            return []
        acl: list[iam.AccessControlRequest] = []
        if not loaded.access_control_list:
            return acl
        for access_control in loaded.access_control_list:
            if not access_control.all_permissions:
                continue
            for permission in access_control.all_permissions:
                if permission.inherited:
                    continue
                acl.append(
                    iam.AccessControlRequest(
                        permission_level=permission.permission_level,
                        service_principal_name=access_control.service_principal_name,
                        group_name=access_control.group_name,
                        user_name=access_control.user_name,
                    )
                )
        # sort to return deterministic results
        return sorted(acl, key=lambda v: f"{v.group_name}:{v.user_name}:{v.service_principal_name}")

    def load_as_dict(self, object_type: str, object_id: str) -> dict[str, iam.PermissionLevel]:
        result = {}
        for acl in self._load_as_request(object_type, object_id):
            if not acl.permission_level:
                continue
            result[self._key_for_acl_dict(acl)] = acl.permission_level
        return result

    @staticmethod
    def _key_for_acl_dict(acl: iam.AccessControlRequest) -> str:
        if acl.group_name is not None:
            return acl.group_name
        if acl.user_name is not None:
            return acl.user_name
        if acl.service_principal_name is not None:
            return acl.service_principal_name
        return "UNKNOWN"

    # TODO remove after ES-892977 is fixed
    @retried(on=[InternalError], timeout=timedelta(minutes=5))
    def _safe_get_permissions(self, object_type: str, object_id: str) -> iam.ObjectPermissions | None:
        try:
            return self._ws.permissions.get(object_type, object_id)
        except PermissionDenied:
            logger.warning(f"permission denied: {object_type} {object_id}")
            return None
        except NotFound:
            logger.warning(f"removed on backend: {object_type} {object_id}")
            return None
        except InvalidParameterValue:
            logger.warning(f"jobs or cluster removed on backend: {object_type} {object_id}")
            return None

    def _safe_update_permissions(
        self, object_type: str, object_id: str, acl: list[iam.AccessControlRequest]
    ) -> iam.ObjectPermissions | None:
        try:
            return self._ws.permissions.update(object_type, object_id, access_control_list=acl)
        except PermissionDenied:
            logger.warning(f"permission denied: {object_type} {object_id}")
            return None
        except NotFound:
            logger.warning(f"removed on backend: {object_type} {object_id}")
            return None
        except InvalidParameterValue:
            logger.warning(f"jobs or cluster removed on backend: {object_type} {object_id}")
            return None

    def _prepare_new_acl(
        self, permissions: iam.ObjectPermissions, migration_state: MigrationState
    ) -> list[iam.AccessControlRequest]:
        _acl = permissions.access_control_list
        acl_requests: list[iam.AccessControlRequest] = []
        if not _acl:
            return acl_requests
        coord = f"{permissions.object_type}/{permissions.object_id}"
        for _item in _acl:
            if not _item.group_name:
                continue
            if not migration_state.is_in_scope(_item.group_name):
                logger.debug(f"Skipping {_item} for {coord} because it is not in scope")
                continue
            new_group_name = migration_state.get_target_principal(_item.group_name)
            if new_group_name is None:
                logger.debug(f"Skipping {_item.group_name} for {coord} because it has no target principal")
                continue
            if not _item.all_permissions:
                continue
            for permission in _item.all_permissions:
                if permission.inherited:
                    continue
                acl_requests.append(
                    iam.AccessControlRequest(
                        group_name=new_group_name,
                        service_principal_name=_item.service_principal_name,
                        user_name=_item.user_name,
                        permission_level=permission.permission_level,
                    )
                )
        return acl_requests

    def __repr__(self):
        return f"GenericPermissionsSupport({self._listings})"


class WorkspaceListing(Listing, CrawlerBase[WorkspaceObjectInfo]):
    def __init__(
        self,
        ws: WorkspaceClient,
        sql_backend: SqlBackend,
        inventory_database: str,
        num_threads=20,
        start_path: str | None = "/",
    ):
        Listing.__init__(self, lambda: [], "_", "_")
        CrawlerBase.__init__(
            self,
            backend=sql_backend,
            catalog="hive_metastore",
            schema=inventory_database,
            table="workspace_objects",
            klass=WorkspaceObjectInfo,
        )
        self._ws = ws
        self._num_threads = num_threads
        self._start_path = start_path
        self._sql_backend = sql_backend
        self._inventory_database = inventory_database

    def _crawl(self) -> Iterable[WorkspaceObjectInfo]:
        # pylint: disable-next=import-outside-toplevel,redefined-outer-name
        from databricks.labs.ucx.workspace_access.listing import WorkspaceListing

        ws_listing = WorkspaceListing(self._ws, num_threads=self._num_threads, with_directories=False)
        for obj in ws_listing.walk(self._start_path):
            if obj is None or obj.object_type is None:
                continue
            raw = obj.as_dict()
            yield WorkspaceObjectInfo(
                object_type=raw.get("object_type", None),
                object_id=str(raw.get("object_id", None)),
                path=raw.get("path", None),
                language=raw.get("language", None),
            )

    def snapshot(self) -> Iterable[WorkspaceObjectInfo]:
        return self._snapshot(self._try_fetch, self._crawl)

    def _try_fetch(self) -> Iterable[WorkspaceObjectInfo]:
        for row in self._fetch(f"SELECT * FROM {escape_sql_identifier(self.full_name)}"):
            yield WorkspaceObjectInfo(
                path=row["path"], object_type=row["object_type"], object_id=row["object_id"], language=row["language"]
            )

    def object_types(self) -> set[str]:
        return {"notebooks", "directories", "repos", "files"}

    @staticmethod
    def _convert_object_type_to_request_type(_object: WorkspaceObjectInfo) -> str | None:
        match _object.object_type:
            case "NOTEBOOK":
                return "notebooks"
            case "DIRECTORY":
                return "directories"
            case "LIBRARY":
                return None
            case "REPO":
                return "repos"
            case "FILE":
                return "files"
        # silent handler for experiments - they'll be inventoried by the experiments manager
        return None

    def __iter__(self):
        for _object in self.snapshot():
            request_type = self._convert_object_type_to_request_type(_object)
            if not request_type:
                continue
            assert _object.object_id is not None
            yield GenericPermissionsInfo(str(_object.object_id), request_type)

    def __repr__(self):
        return f"WorkspaceListing(start_path={self._start_path})"


def models_listing(ws: WorkspaceClient, num_threads: int):
    def inner() -> Iterator[ml.ModelDatabricks]:
        tasks = []
        for model in ws.model_registry.list_models():
            tasks.append(partial(ws.model_registry.get_model, name=model.name))
        models, errors = Threads.gather("listing model ids", tasks, num_threads)
        if len(errors) > 0:
            raise ManyError(errors)
        for model_response in models:
            if not model_response.registered_model_databricks:
                continue
            yield model_response.registered_model_databricks

    return inner


def experiments_listing(ws: WorkspaceClient):
    def inner() -> Iterator[ml.Experiment]:
        for experiment in ws.experiments.list_experiments():
            # We filter-out notebook-based experiments, because they are covered by notebooks listing in
            # workspace-based notebook experiment
            if experiment.tags:
                nb_tag = [t for t in experiment.tags if t.key == "mlflow.experimentType" and t.value == "NOTEBOOK"]
                # repo-based notebook experiment
                repo_nb_tag = [
                    t for t in experiment.tags if t.key == "mlflow.experiment.sourceType" and t.value == "REPO_NOTEBOOK"
                ]
                if nb_tag or repo_nb_tag:
                    continue

            yield experiment

    return inner


def feature_store_listing(ws: WorkspaceClient):
    def inner() -> list[GenericPermissionsInfo]:
        feature_tables = []
        token = None
        while True:
            result = ws.api_client.do(
                "GET", "/api/2.0/feature-store/feature-tables/search", query={"page_token": token, "max_results": 200}
            )
            assert isinstance(result, dict)
            for table in result.get("feature_tables", []):
                feature_tables.append(GenericPermissionsInfo(table["id"], "feature-tables"))

            if "next_page_token" not in result:
                break
            token = result["next_page_token"]  # type: ignore[index]

        return feature_tables

    return inner


def feature_tables_root_page():
    return [GenericPermissionsInfo("/root", "feature-tables")]


def models_root_page():
    return [GenericPermissionsInfo("/root", "registered-models")]


def tokens_and_passwords():
    for _value in ("tokens", "passwords"):
        yield GenericPermissionsInfo(_value, "authorization")



import logging
from urllib.parse import urlparse

from databricks.sdk import WorkspaceClient
from databricks.sdk.errors.platform import InvalidParameterValue, PermissionDenied
from databricks.labs.ucx.azure.access import AzureResourcePermissions
from databricks.labs.ucx.azure.resources import AzureResources
from databricks.labs.ucx.hive_metastore import ExternalLocations
from databricks.labs.ucx.hive_metastore.grants import PrincipalACL


logger = logging.getLogger(__name__)


class ExternalLocationsMigration:
    def __init__(
        self,
        ws: WorkspaceClient,
        hms_locations: ExternalLocations,
        resource_permissions: AzureResourcePermissions,
        azurerm: AzureResources,
        principal_acl: PrincipalACL,
    ):
        self._ws = ws
        self._hms_locations = hms_locations
        self._resource_permissions = resource_permissions
        self._azurerm = azurerm
        self._principal_acl = principal_acl

    def _app_id_credential_name_mapping(self) -> tuple[dict[str, str], dict[str, str]]:
        # list all storage credentials.
        # generate the managed identity/service principal application id to credential name mapping.
        # return one mapping for all non read-only credentials and one mapping for all read-only credentials
        # TODO: considering put this logic into the StorageCredentialManager
        app_id_mapping_write = {}
        app_id_mapping_read = {}
        all_credentials = self._ws.storage_credentials.list(max_results=0)
        for credential in all_credentials:
            name = credential.name
            # cannot have none credential name, it's required for external location
            if not name:
                continue

            read_only = credential.read_only
            service_principal = credential.azure_service_principal
            managed_identity = credential.azure_managed_identity

            application_id = None
            if service_principal:
                # if service principal based credential, use service principal's application_id directly
                application_id = service_principal.application_id
            if managed_identity:
                # if managed identity based credential, fetch the application_id of the managed identity
                application_id = self._azurerm.managed_identity_client_id(
                    managed_identity.access_connector_id,
                    managed_identity.managed_identity_id,
                )
            if not application_id:
                continue

            if read_only:
                app_id_mapping_read[application_id] = name
                continue
            app_id_mapping_write[application_id] = name

        return app_id_mapping_write, app_id_mapping_read

    def _prefix_credential_name_mapping(self) -> tuple[dict[str, str], dict[str, str]]:
        # get managed identity/service principal's application id to storage credential name mapping
        # for all non read-only and read-only credentials
        app_id_mapping_write, app_id_mapping_read = self._app_id_credential_name_mapping()

        # use the application id to storage credential name mapping to create prefix to storage credential name mapping
        prefix_mapping_write = {}
        prefix_mapping_read = {}
        for permission_mapping in self._resource_permissions.load():
            if permission_mapping.client_id in app_id_mapping_write:
                prefix_mapping_write[permission_mapping.prefix] = app_id_mapping_write[permission_mapping.client_id]
                continue
            if permission_mapping.client_id in app_id_mapping_read:
                prefix_mapping_read[permission_mapping.prefix] = app_id_mapping_read[permission_mapping.client_id]

        all_storage_accounts = list(self._azurerm.storage_accounts())
        for storage_credential in self._ws.storage_credentials.list():
            # Filter storage credentials for access connectors created by UCX
            if not (
                storage_credential.name is not None
                and storage_credential.name.startswith("ac-")
                and storage_credential.comment is not None
                and storage_credential.comment == "Created by UCX"
            ):
                continue

            storage_account_name = storage_credential.name.removeprefix("ac-")
            storage_accounts = [st for st in all_storage_accounts if st.name == storage_account_name]
            if len(storage_accounts) == 0:
                logger.warning(
                    f"Storage account {storage_account_name} for access connector {storage_credential.name} not found, "
                    "therefore, not able to create external locations for this storage account using the access "
                    "connector."
                )
                continue

            for container in self._azurerm.containers(storage_accounts[0].id):
                storage_url = f"abfss://{container.container}@{container.storage_account}.dfs.core.windows.net/"
                # UCX assigns created access connectors the "STORAGE_BLOB_DATA_CONTRIBUTOR" role on the storage account
                prefix_mapping_write[storage_url] = storage_credential.name

        return prefix_mapping_write, prefix_mapping_read

    def _create_location_name(self, location_url: str) -> str:
        # generate the UC external location name
        before_at, _, after_at = location_url.partition('@')
        container_name = before_at.removeprefix("abfss://")
        res_name = after_at.replace(".dfs.core.windows.net", "").rstrip("/").replace("/", "_")
        return f"{container_name}_{res_name}"

    def _create_external_location_helper(
        self, name, url, credential, comment="Created by UCX", read_only=False, skip_validation=False
    ) -> str | None:
        try:
            self._ws.external_locations.create(
                name, url, credential, comment=comment, read_only=read_only, skip_validation=skip_validation
            )
            return url
        except InvalidParameterValue as invalid:
            if "overlaps with an existing external location" in str(invalid):
                logger.warning(f"Skip creating external location, see details: {str(invalid)}")
                return None
            raise invalid

    def _create_external_location(
        self, location_url: str, prefix_mapping_write: dict[str, str], prefix_mapping_read: dict[str, str]
    ) -> str | None:
        location_name = self._create_location_name(location_url)

        # get container url as the prefix
        parsed_url = urlparse(location_url)
        container_url = f"{parsed_url.scheme}://{parsed_url.netloc}/"

        # try to create external location with write privilege first
        if container_url in prefix_mapping_write:
            url = self._create_external_location_helper(
                location_name, location_url, prefix_mapping_write[container_url], comment="Created by UCX"
            )
            return url
        # if no matched write privilege credential, try to create read-only external location
        if container_url in prefix_mapping_read:
            try:
                url = self._create_external_location_helper(
                    location_name,
                    location_url,
                    prefix_mapping_read[container_url],
                    comment="Created by UCX",
                    read_only=True,
                )
                return url
            except PermissionDenied as denied:
                if "No file available under the location to read" in str(denied):
                    # Empty location will cause failed READ permission check with read-only credential
                    # Skip skip_validation in this case
                    url = self._create_external_location_helper(
                        location_name,
                        location_url,
                        prefix_mapping_read[container_url],
                        comment="Created by UCX",
                        read_only=True,
                        skip_validation=True,
                    )
                    return url
                raise denied
        # if no credential found
        return None

    def _filter_unsupported_location(self, location_urls: list[str]) -> list[str]:
        # remove unsupported external location
        supported_urls = []
        for url in location_urls:
            if url.startswith("abfss://"):
                supported_urls.append(url)
                continue
            logger.warning(f"Skip unsupported location: {url}")
        return supported_urls

    def run(self):
        # list missing external locations in UC
        _, missing_locations = self._hms_locations.match_table_external_locations()
        # Extract the location URLs from the missing locations
        missing_loc_urls = [loc.location for loc in missing_locations]
        missing_loc_urls = self._filter_unsupported_location(missing_loc_urls)

        # get prefix to storage credential name mapping
        prefix_mapping_write, prefix_mapping_read = self._prefix_credential_name_mapping()

        # if missing external location is in prefix to storage credential name mapping
        # create a UC external location with mapped storage credential name
        migrated_loc_urls = []
        for location_url in missing_loc_urls:
            migrated_loc_url = self._create_external_location(location_url, prefix_mapping_write, prefix_mapping_read)
            if migrated_loc_url:
                migrated_loc_urls.append(migrated_loc_url)

        leftover_loc_urls = [url for url in missing_loc_urls if url not in migrated_loc_urls]
        self._principal_acl.apply_location_acl()
        if leftover_loc_urls:
            logger.info(
                "External locations below are not created in UC. You may check following cases and rerun this command:"
                "1. Please check the output of 'migrate_credentials' command for storage credentials migration failure."
                "2. If you use service principal in extra_config when create dbfs mount or use service principal "
                "in your code directly for storage access, UCX cannot automatically migrate them to storage credential."
                "Please manually create those storage credentials first."
                "3. You may have overlapping external location already in UC."
            )
            for loc_url in leftover_loc_urls:
                logger.info(f"Not created external location: {loc_url}")
            return leftover_loc_urls

        logger.info("All UC external location are created.")
        return leftover_loc_urls



import logging

from databricks.labs.ucx.assessment.workflows import Assessment
from databricks.labs.ucx.contexts.workflow_task import RuntimeContext
from databricks.labs.ucx.framework.tasks import Workflow, job_task

logger = logging.getLogger(__name__)


class GroupMigration(Workflow):
    def __init__(self):
        super().__init__('migrate-groups')

    @job_task(job_cluster="table_migration")
    def verify_metastore_attached(self, ctx: RuntimeContext):
        """Verifies if a metastore is attached to this workspace. If not, the workflow will fail."""
        ctx.verify_has_metastore.verify_metastore()

    @job_task(depends_on=[Assessment.crawl_groups, verify_metastore_attached])
    def rename_workspace_local_groups(self, ctx: RuntimeContext):
        """Renames workspace local groups by adding `db-temp-` prefix."""
        ctx.group_manager.rename_groups()

    @job_task(depends_on=[rename_workspace_local_groups])
    def reflect_account_groups_on_workspace(self, ctx: RuntimeContext):
        """Adds matching account groups to this workspace. The matching account level group(s) must preexist(s) for this
        step to be successful. This process does not create the account level group(s)."""
        ctx.group_manager.reflect_account_groups_on_workspace()

    @job_task(depends_on=[reflect_account_groups_on_workspace], job_cluster="tacl")
    def apply_permissions_to_account_groups(self, ctx: RuntimeContext):
        """Fourth phase of the workspace-local group migration process. It does the following:
          - Assigns the full set of permissions of the original group to the account-level one

        It covers local workspace-local permissions for all entities: Legacy Table ACLs, Entitlements,
        AWS instance profiles, Clusters, Cluster policies, Instance Pools, Databricks SQL warehouses, Delta Live
        Tables, Jobs, MLflow experiments, MLflow registry, SQL Dashboards & Queries, SQL Alerts, Token and Password usage
        permissions, Secret Scopes, Notebooks, Directories, Repos, Files.

        See [interactive tutorial here](https://app.getreprise.com/launch/myM3VNn/)."""
        migration_state = ctx.group_manager.get_migration_state()
        if len(migration_state.groups) == 0:
            logger.info("Skipping group migration as no groups were found.")
            return
        ctx.permission_manager.apply_group_permissions(migration_state)

    @job_task(depends_on=[apply_permissions_to_account_groups], job_cluster="tacl")
    def validate_groups_permissions(self, ctx: RuntimeContext):
        """Validate that all the crawled permissions are applied correctly to the destination groups."""
        ctx.permission_manager.verify_group_permissions()


class PermissionsMigrationAPI(Workflow):
    def __init__(self):
        super().__init__('migrate-groups-experimental')

    @job_task(job_cluster="table_migration")
    def verify_metastore_attached(self, ctx: RuntimeContext):
        """Verifies if a metastore is attached to this workspace. If not, the workflow will fail."""
        ctx.verify_has_metastore.verify_metastore()

    @job_task(depends_on=[Assessment.crawl_groups, verify_metastore_attached])
    def rename_workspace_local_groups(self, ctx: RuntimeContext):
        """[EXPERIMENTAL] Renames workspace local groups by adding `db-temp-` prefix."""
        ctx.group_manager.rename_groups()

    @job_task(depends_on=[rename_workspace_local_groups])
    def reflect_account_groups_on_workspace(self, ctx: RuntimeContext):
        """[EXPERIMENTAL] Adds matching account groups to this workspace. The matching account level group(s) must preexist(s) for this
        step to be successful. This process does not create the account level group(s)."""
        ctx.group_manager.reflect_account_groups_on_workspace()

    @job_task(depends_on=[reflect_account_groups_on_workspace])
    def apply_permissions(self, ctx: RuntimeContext):
        """[EXPERIMENTAL] This task uses the new permission migration API which requires enrolment from Databricks.
        Fourth phase of the workspace-local group migration process. It does the following:
          - Assigns the full set of permissions of the original group to the account-level one

        The permission migration is not atomic. If we hit InternalError, it is possible that half the permissions
        have already been migrated over to the account group, and the other half of the permissions are still with
        the workspace local group. Addressing cases like this would require two options that both require manual
        intervention:
          - Deleting all conflicting permissions for the account group, rerun the permission migration between
            the workspace and account group (the workflow is idempotent).
          - Creating a new account group, reverting all the permissions that were migrated over to the old account
            group, and running the workflow again.

        To make things run smoothly, this workflow should never be run on an account group that already has permissions.
        The expectation is that account group has no permissions to begin with.

        It covers local workspace-local permissions for all entities."""
        migration_state = ctx.group_manager.get_migration_state()
        if len(migration_state.groups) == 0:
            logger.info("Skipping group migration as no groups were found.")
        elif migration_state.apply_to_renamed_groups(ctx.workspace_client):
            logger.info("Group permission migration completed successfully.")
        else:
            msg = "Permission migration for groups failed; reason unknown."
            raise RuntimeError(msg)


class ValidateGroupPermissions(Workflow):
    def __init__(self):
        super().__init__('validate-groups-permissions')

    @job_task(job_cluster="tacl")
    def validate_groups_permissions(self, ctx: RuntimeContext):
        """Validate that all the crawled permissions are applied correctly to the destination groups."""
        ctx.permission_manager.verify_group_permissions()


class RemoveWorkspaceLocalGroups(Workflow):
    def __init__(self):
        super().__init__('remove-workspace-local-backup-groups')

    @job_task(depends_on=[GroupMigration.apply_permissions_to_account_groups])
    def delete_backup_groups(self, ctx: RuntimeContext):
        """Last step of the group migration process. Removes all workspace-level backup groups, along with their
        permissions. Execute this workflow only after you've confirmed that workspace-local migration worked
        successfully for all the groups involved."""
        ctx.group_manager.delete_original_workspace_groups()



# pylint: disable=duplicate-code
import dataclasses
import json
import logging
from collections.abc import Callable, Iterable
from dataclasses import dataclass
from datetime import timedelta
from functools import partial

from databricks.labs.blueprint.limiter import rate_limited
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import InternalError, NotFound, PermissionDenied
from databricks.sdk.retries import retried
from databricks.sdk.service import sql
from databricks.sdk.service.sql import ObjectTypePlural, SetResponse

from databricks.labs.ucx.workspace_access.base import AclSupport, Permissions, StaticListing
from databricks.labs.ucx.workspace_access.groups import MigrationState

logger = logging.getLogger(__name__)


@dataclass
class SqlPermissionsInfo:
    object_id: str
    request_type: sql.ObjectTypePlural


# This module is called redash to disambiguate from databricks.sdk.service.sql


class Listing:
    def __init__(self, func: Callable[..., Iterable], request_type: sql.ObjectTypePlural):
        self._func = func
        self._request_type = request_type
        self.object_type = request_type.value

    def __iter__(self):
        for item in self._func():
            yield SqlPermissionsInfo(item.id, self._request_type)

    def __repr__(self):
        return f"Listing({self.object_type} via {self._func.__qualname__})"


class RedashPermissionsSupport(AclSupport):
    def __init__(
        self,
        ws: WorkspaceClient,
        listings: list[Listing],
        set_permissions_timeout: timedelta | None = timedelta(minutes=1),
        # Group information in Redash are cached for up to 10 minutes causing inconsistencies.
        # If a group is renamed, the old name may still be returned by the dbsql get permissions api.
        # Note that the update/set API is strongly consistent and is not affected by this behaviour.
        # The validation step should keep retrying for at least 10 mins until the get api returns the new group name.
        # More details here: https://databricks.atlassian.net/browse/ES-992619
        verify_timeout: timedelta | None = timedelta(minutes=11),
        # this parameter is for testing scenarios only - [{object_type}:{object_id}]
        # it will use StaticListing class to return only object ids that has the same object type
        include_object_permissions: list[str] | None = None,
    ):
        self._ws = ws
        self._listings = listings
        self._set_permissions_timeout = set_permissions_timeout
        self._verify_timeout = verify_timeout
        self._include_object_permissions = include_object_permissions

    @staticmethod
    def _is_item_relevant(item: Permissions, migration_state: MigrationState) -> bool:
        permissions_response = sql.GetResponse.from_dict(json.loads(item.raw))
        assert permissions_response.access_control_list is not None
        mentioned_groups = [acl.group_name for acl in permissions_response.access_control_list]
        return any(g in mentioned_groups for g in [info.name_in_workspace for info in migration_state.groups])

    def get_crawler_tasks(self):
        if self._include_object_permissions:
            for item in StaticListing(self._include_object_permissions, self.object_types()):
                yield partial(self._crawler_task, item.object_id, sql.ObjectTypePlural(item.object_type))
            return
        for listing in self._listings:
            for item in listing:
                yield partial(self._crawler_task, item.object_id, item.request_type)

    def object_types(self) -> set[str]:
        all_object_types = set()
        for listing in self._listings:
            all_object_types.add(listing.object_type)
        return all_object_types

    def get_apply_task(self, item: Permissions, migration_state: MigrationState):
        if not self._is_item_relevant(item, migration_state):
            return None
        permissions = sql.GetResponse.from_dict(json.loads(item.raw))
        assert permissions.access_control_list is not None
        new_acl = self._prepare_new_acl(permissions.access_control_list, migration_state)
        return partial(
            self._applier_task,
            object_type=sql.ObjectTypePlural(item.object_type),
            object_id=item.object_id,
            acl=new_acl,
        )

    def _safe_get_dbsql_permissions(self, object_type: sql.ObjectTypePlural, object_id: str) -> sql.GetResponse | None:
        try:
            return self._ws.dbsql_permissions.get(object_type, object_id)
        except NotFound:
            logger.warning(f"removed on backend: {object_type} {object_id}")
            return None

    def _load_as_request(self, object_type: sql.ObjectTypePlural, object_id: str) -> list[sql.AccessControl]:
        loaded = self._safe_get_dbsql_permissions(object_type, object_id)
        if loaded is None:
            return []
        acl: list[sql.AccessControl] = []
        if not loaded.access_control_list:
            return acl

        for access_control in loaded.access_control_list:
            acl.append(
                sql.AccessControl(
                    permission_level=access_control.permission_level,
                    group_name=access_control.group_name,
                    user_name=access_control.user_name,
                )
            )
        # sort to return deterministic results
        return sorted(acl, key=lambda v: f"{v.group_name}:{v.user_name}")

    def load_as_dict(self, object_type: sql.ObjectTypePlural, object_id: str) -> dict[str, sql.PermissionLevel]:
        result = {}
        for acl in self._load_as_request(object_type, object_id):
            if not acl.permission_level:
                continue
            result[self._key_for_acl_dict(acl)] = acl.permission_level
        return result

    @staticmethod
    def _key_for_acl_dict(acl: sql.AccessControl) -> str:
        if acl.group_name is not None:
            return acl.group_name
        if acl.user_name is not None:
            return acl.user_name
        return "UNKNOWN"

    @rate_limited(max_requests=100)
    def _crawler_task(self, object_id: str, object_type: sql.ObjectTypePlural) -> Permissions | None:
        permissions = self._safe_get_dbsql_permissions(object_type=object_type, object_id=object_id)
        if permissions:
            return Permissions(
                object_id=object_id,
                object_type=object_type.value,
                raw=json.dumps(permissions.as_dict()),
            )
        return None

    @rate_limited(max_requests=100)
    def _verify(self, object_type: sql.ObjectTypePlural, object_id: str, acl: list[sql.AccessControl]):
        # in-flight check for the applied permissions
        # the api might be inconsistent, therefore we need to check that the permissions were applied
        remote_permission = self._safe_get_dbsql_permissions(object_type, object_id)
        if remote_permission:
            assert remote_permission.access_control_list is not None
            if all(elem in remote_permission.access_control_list for elem in acl):
                return True
            msg = (
                f"Couldn't find permission for object type {object_type} with id {object_id}\n"
                f"acl to be applied={acl}\n"
                f"acl found in the object={remote_permission}\n"
            )
            raise NotFound(msg)
        return False

    def get_verify_task(self, item: Permissions) -> Callable[[], bool]:
        acl = sql.GetResponse.from_dict(json.loads(item.raw))
        if not acl.access_control_list:
            raise ValueError(
                f"Access control list not present for object type " f"{item.object_type} and object id {item.object_id}"
            )
        return partial(self._verify, sql.ObjectTypePlural(item.object_type), item.object_id, acl.access_control_list)

    @rate_limited(max_requests=30)
    def _applier_task(self, object_type: sql.ObjectTypePlural, object_id: str, acl: list[sql.AccessControl]):
        """
        Please note that we only have SET option (DBSQL Permissions API doesn't support UPDATE operation).
        This affects the way how we prepare the new ACL request.
        """

        set_retry_on_value_error = retried(on=[InternalError, ValueError], timeout=self._set_permissions_timeout)
        set_retried_check = set_retry_on_value_error(self._safe_set_permissions)
        set_retried_check(object_type, object_id, acl)

        retry_on_value_error = retried(on=[InternalError, NotFound], timeout=self._verify_timeout)
        retried_check = retry_on_value_error(self._verify)
        return retried_check(object_type, object_id, acl)

    @staticmethod
    def _prepare_new_acl(acl: list[sql.AccessControl], migration_state: MigrationState) -> list[sql.AccessControl]:
        """
        Please note the comment above on how we apply these permissions.
        Permissions are set/replaced and not updated/patched, therefore all existing ACLs need to be collected
        including users and temp/backup groups.
        """
        acl_requests: list[sql.AccessControl] = []
        for access_control in acl:
            if access_control.user_name:
                logger.debug(f"Including redash permissions acl for user: `{access_control.user_name}`")
                acl_requests.append(access_control)
                continue

            if not access_control.group_name:
                continue

            if not migration_state.is_in_scope(access_control.group_name):
                logger.debug(
                    f"Including redash permissions acl for group not in the scope: `{access_control.group_name}`"
                )
                acl_requests.append(access_control)
                continue

            target_principal = migration_state.get_target_principal(access_control.group_name)
            if not target_principal:
                logger.debug(
                    f"Including redash permissions acl for group without target principal: "
                    f"`{access_control.group_name}`"
                )
                acl_requests.append(access_control)
                continue

            logger.debug(f"Including redash permissions acl for target group `{target_principal}`")
            new_acl_request = dataclasses.replace(access_control, group_name=target_principal)
            acl_requests.append(new_acl_request)

            temp_principal = migration_state.get_temp_principal(access_control.group_name)
            if temp_principal is not None:
                logger.debug(f"Including redash permissions acl for temp group `{temp_principal}`")
                temp_group_new_acl_request = dataclasses.replace(access_control, group_name=temp_principal)
                acl_requests.append(temp_group_new_acl_request)

        return acl_requests

    @rate_limited(burst_period_seconds=30)
    def _safe_set_permissions(
        self, object_type: ObjectTypePlural, object_id: str, acl: list[sql.AccessControl] | None
    ) -> SetResponse | None:
        assert acl is not None

        def hash_permissions(permissions: list[sql.AccessControl]):
            return {
                hash((permission.permission_level, permission.user_name, permission.group_name))
                for permission in permissions
            }

        try:
            res = self._ws.dbsql_permissions.set(object_type=object_type, object_id=object_id, access_control_list=acl)
            assert res.access_control_list is not None
            if hash_permissions(acl).issubset(hash_permissions(res.access_control_list)):
                return res
            msg = (
                f"Failed to set permission and will be retried for {object_type} {object_id}, "
                f"doing another attempt..."
            )
            raise ValueError(msg)
        except PermissionDenied:
            logger.warning(f"Permission denied: {object_type} {object_id}")
            return None
        except NotFound:
            logger.warning(f"Deleted on platform: {object_type} {object_id}")
            return None

    def __repr__(self):
        return f"RedashPermissionsSupport({self._listings})"


def redash_listing_wrapper(
    func: Callable[..., list], object_type: sql.ObjectTypePlural
) -> Callable[..., Iterable[SqlPermissionsInfo]]:
    def wrapper() -> Iterable[SqlPermissionsInfo]:
        for item in func():
            yield SqlPermissionsInfo(
                object_id=item.id,
                request_type=object_type,
            )

    return wrapper



import json
import logging
import re
import shutil
import typing
from collections.abc import Callable
from dataclasses import dataclass, field
from datetime import timedelta

from databricks.sdk.errors import NotFound
from databricks.sdk.retries import retried
from databricks.sdk.service.catalog import Privilege

from databricks.labs.ucx.framework.utils import run_command

logger = logging.getLogger(__name__)


@dataclass
class AWSRole:
    path: str
    role_name: str
    role_id: str
    arn: str


@dataclass
class AWSPolicyAction:
    resource_type: str
    privilege: str
    resource_path: str


@dataclass
class AWSUCRoleCandidate:
    """Candidates for UC IAM roles with the paths they have access to"""

    role_name: str
    policy_name: str
    resource_paths: list[str]


@dataclass
class AWSRoleAction:
    role_arn: str
    resource_type: str
    privilege: str
    resource_path: str

    @property
    def role_name(self):
        role_match = re.match(AWSResources.ROLE_NAME_REGEX, self.role_arn)
        return role_match.group(1)


@dataclass
class AWSInstanceProfile:
    instance_profile_arn: str
    iam_role_arn: str | None = None

    @property
    def role_name(self) -> str | None:
        if self.iam_role_arn:
            arn = self.iam_role_arn
        else:
            arn = self.instance_profile_arn
        role_match = re.match(AWSResources.ROLE_NAME_REGEX, arn)
        if not role_match:
            logger.error(f"Role ARN is mismatched {self.iam_role_arn}")
            return None
        return role_match.group(1)


@dataclass()
class AWSCredentialCandidate:
    role_arn: str
    privilege: str
    paths: set[str] = field(default_factory=set)

    @property
    def role_name(self):
        role_match = re.match(AWSResources.ROLE_NAME_REGEX, self.role_arn)
        return role_match.group(1)


class AWSResources:
    S3_ACTIONS: typing.ClassVar[set[str]] = {"s3:PutObject", "s3:GetObject", "s3:DeleteObject", "s3:PutObjectAcl"}
    S3_READONLY: typing.ClassVar[str] = "s3:GetObject"
    S3_REGEX: typing.ClassVar[str] = r"arn:aws:s3:::([a-zA-Z0-9\/+=,.@_-]*)\/\*$"
    S3_BUCKET: typing.ClassVar[str] = r"((s3:\/\/|s3a:\/\/)([a-zA-Z0-9+=,.@_-]*))\/.*$"
    S3_PREFIX: typing.ClassVar[str] = "arn:aws:s3:::"
    S3_PATH_REGEX: typing.ClassVar[str] = r"((s3:\/\/)|(s3a:\/\/))(.*)"
    UC_MASTER_ROLES_ARN: typing.ClassVar[list[str]] = [
        "arn:aws:iam::414351767826:role/unity-catalog-prod-UCMasterRole-14S5ZJVKOTYTL",
        "arn:aws:iam::707343435239:role/unity-catalog-dev-UCMasterRole-G3MMN8SP21FO",
    ]
    ROLE_NAME_REGEX = r"arn:aws:iam::[0-9]+:(?:instance-profile|role)\/([a-zA-Z0-9+=,.@_-]*)$"

    def __init__(self, profile: str, command_runner: Callable[[str], tuple[int, str, str]] = run_command):
        self._profile = profile
        self._command_runner = command_runner

    def validate_connection(self):
        validate_command = "sts get-caller-identity"
        result = self._run_json_command(validate_command)
        if result:
            logger.info(result)
            return result
        return None

    def list_role_policies(self, role_name: str):
        list_policies_cmd = f"iam list-role-policies --role-name {role_name}"
        policies = self._run_json_command(list_policies_cmd)
        if not policies:
            return []
        return policies.get("PolicyNames", [])

    def list_attached_policies_in_role(self, role_name: str):
        list_attached_policies_cmd = f"iam list-attached-role-policies --role-name {role_name}"
        policies = self._run_json_command(list_attached_policies_cmd)
        if not policies:
            return []
        attached_policies = []
        for policy in policies.get("AttachedPolicies", []):
            attached_policies.append(policy.get("PolicyArn"))
        return attached_policies

    def list_all_uc_roles(self) -> list[AWSRole]:
        roles = self._run_json_command("iam list-roles")
        uc_roles: list[AWSRole] = []
        roles = roles.get("Roles")
        if not roles:
            logger.warning("list-roles couldn't find any roles")
            return uc_roles
        for role in roles:
            policy_document = role.get("AssumeRolePolicyDocument")
            if not policy_document:
                continue
            for statement in policy_document["Statement"]:
                effect = statement.get("Effect")
                action = statement.get("Action")
                principal = statement.get("Principal")
                if not (effect and action and principal):
                    continue
                if effect != "Allow":
                    continue
                if action != "sts:AssumeRole":
                    continue
                principal = principal.get("AWS")
                if not principal:
                    continue
                if not self._is_uc_principal(principal):
                    continue
                uc_roles.append(
                    AWSRole(
                        role_id=role.get("RoleId"),
                        role_name=role.get("RoleName"),
                        arn=role.get("Arn"),
                        path=role.get("Path"),
                    )
                )

        return uc_roles

    def _is_uc_principal(self, principal):
        if isinstance(principal, list):
            for single_principal in principal:
                if single_principal in self.UC_MASTER_ROLES_ARN:
                    return True
            return False
        return principal in self.UC_MASTER_ROLES_ARN

    def get_role_policy(self, role_name, policy_name: str | None = None, attached_policy_arn: str | None = None):
        if policy_name:
            get_policy = f"iam get-role-policy --role-name {role_name} --policy-name {policy_name}"
        elif attached_policy_arn:
            get_attached_policy = f"iam get-policy --policy-arn {attached_policy_arn}"
            attached_policy = self._run_json_command(get_attached_policy)
            if not attached_policy:
                return []
            policy_version = attached_policy["Policy"]["DefaultVersionId"]
            get_policy = f"iam get-policy-version --policy-arn {attached_policy_arn} --version-id {policy_version}"
        else:
            logger.error("Failed to retrieve role. No role name or attached role ARN specified.")
            return []
        policy = self._run_json_command(get_policy)
        if not policy:
            return []
        if policy_name:
            actions = policy["PolicyDocument"].get("Statement", [])
        else:
            actions = policy["PolicyVersion"]["Document"].get("Statement", [])
        return self._policy_actions(actions)

    def _policy_actions(self, actions):
        policy_actions = []
        for action in actions:
            if action.get("Effect", "Deny") != "Allow":
                continue
            actions = action["Action"]
            s3_actions = self._s3_actions(actions)
            if not s3_actions or self.S3_READONLY not in s3_actions:
                continue
            privilege = Privilege.WRITE_FILES.value
            for s3_action_type in self.S3_ACTIONS:
                if s3_action_type not in s3_actions:
                    privilege = Privilege.READ_FILES.value
                    continue
            for resource in action.get("Resource", []):
                match = re.match(self.S3_REGEX, resource)
                if match:
                    policy_actions.append(AWSPolicyAction("s3", privilege, f"s3://{match.group(1)}"))
                    policy_actions.append(AWSPolicyAction("s3", privilege, f"s3a://{match.group(1)}"))
        return policy_actions

    def _s3_actions(self, actions):
        s3_actions = []
        if isinstance(actions, list):
            for single_action in actions:
                if single_action in self.S3_ACTIONS:
                    s3_actions.append(single_action)
                    continue
        elif actions in self.S3_ACTIONS:
            s3_actions = [actions]
        return s3_actions

    def _aws_role_trust_doc(self, self_assume_arn: str | None = None, external_id="0000"):
        return self._get_json_for_cli(
            {
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Effect": "Allow",
                        "Principal": {
                            "AWS": (
                                [
                                    "arn:aws:iam::414351767826:role/unity-catalog-prod-UCMasterRole-14S5ZJVKOTYTL",
                                    self_assume_arn,
                                ]
                                if self_assume_arn
                                else "arn:aws:iam::414351767826:role/unity-catalog-prod-UCMasterRole-14S5ZJVKOTYTL"
                            )
                        },
                        "Action": "sts:AssumeRole",
                        "Condition": {"StringEquals": {"sts:ExternalId": external_id}},
                    }
                ],
            }
        )

    def _aws_s3_policy(self, s3_prefixes, account_id, role_name, kms_key=None):
        """
        Create the UC IAM policy for the given S3 prefixes, account ID, role name, and KMS key.
        """
        s3_prefixes_strip = set()
        for path in s3_prefixes:
            match = re.match(AWSResources.S3_PATH_REGEX, path)
            if match:
                s3_prefixes_strip.add(match.group(4))

        s3_prefixes_enriched = sorted([self.S3_PREFIX + s3_prefix for s3_prefix in s3_prefixes_strip])
        statement = [
            {
                "Action": [
                    "s3:GetObject",
                    "s3:PutObject",
                    "s3:PutObjectAcl",
                    "s3:DeleteObject",
                    "s3:ListBucket",
                    "s3:GetBucketLocation",
                ],
                "Resource": s3_prefixes_enriched,
                "Effect": "Allow",
            },
            {
                "Action": ["sts:AssumeRole"],
                "Resource": [f"arn:aws:iam::{account_id}:role/{role_name}"],
                "Effect": "Allow",
            },
        ]
        if kms_key:
            statement.append(
                {
                    "Action": ["kms:Decrypt", "kms:Encrypt", "kms:GenerateDataKey*"],
                    "Resource": [f"arn:aws:kms:{kms_key}"],
                    "Effect": "Allow",
                }
            )
        return self._get_json_for_cli(
            {
                "Version": "2012-10-17",
                "Statement": statement,
            }
        )

    def _create_role(self, role_name: str, assume_role_json: str) -> str | None:
        """
        Create an AWS role with the given name and assume role policy document.
        """
        add_role = self._run_json_command(
            f"iam create-role --role-name {role_name} --assume-role-policy-document {assume_role_json}"
        )
        if not add_role:
            return None
        return add_role["Role"]["Arn"]

    def _update_role(self, role_name: str, assume_role_json: str) -> str | None:
        """
        Create an AWS role with the given name and assume role policy document.
        """
        update_role = self._run_json_command(
            f"iam update-assume-role-policy --role-name {role_name} --policy-document {assume_role_json}"
        )
        return update_role

    def create_uc_role(self, role_name: str) -> str | None:
        """
        Create an IAM role for Unity Catalog to access the S3 buckets.
        the AssumeRole condition will be modified later with the external ID captured from the UC credential.
        https://docs.databricks.com/en/connect/unity-catalog/storage-credentials.html
        """
        return self._create_role(role_name, self._aws_role_trust_doc())

    @retried(on=[NotFound], timeout=timedelta(seconds=30))
    def update_uc_role(self, role_name: str, role_arn: str, external_id: str = "0000") -> str | None:
        """
        Create an IAM role for Unity Catalog to access the S3 buckets.
        the AssumeRole condition will be modified later with the external ID captured from the UC credential.
        https://docs.databricks.com/en/connect/unity-catalog/storage-credentials.html
        """
        result = self._update_role(role_name, self._aws_role_trust_doc(role_arn, external_id))
        logger.debug(f"Updated role {role_name} with {result}")
        if result is None:
            raise NotFound("Assume role policy not updated.")
        return result

    def put_role_policy(
        self,
        role_name: str,
        policy_name: str,
        s3_prefixes: set[str],
        account_id: str,
        kms_key=None,
    ) -> bool:
        if not self._run_command(
            f"iam put-role-policy --role-name {role_name} --policy-name {policy_name} "
            f"--policy-document {self._aws_s3_policy(s3_prefixes, account_id, role_name, kms_key)}"
        ):
            return False
        return True

    def create_migration_role(self, role_name: str) -> str | None:
        aws_role_trust_doc = {
            "Version": "2012-10-17",
            "Statement": [
                {"Effect": "Allow", "Principal": {"Service": "ec2.amazonaws.com"}, "Action": "sts:AssumeRole"}
            ],
        }
        assume_role_json = self._get_json_for_cli(aws_role_trust_doc)
        return self._create_role(role_name, assume_role_json)

    def get_instance_profile_arn(self, instance_profile_name: str) -> str | None:
        instance_profile = self._run_json_command(
            f"iam get-instance-profile --instance-profile-name {instance_profile_name}"
        )

        if not instance_profile:
            return None

        return instance_profile["InstanceProfile"]["Arn"]

    def get_instance_profile_role_arn(self, instance_profile_name: str) -> str | None:
        instance_profile = self._run_json_command(
            f"iam get-instance-profile --instance-profile-name {instance_profile_name}"
        )

        if not instance_profile:
            return None

        try:
            return instance_profile["InstanceProfile"]["Roles"][0]["Arn"]
        except (KeyError, IndexError):
            return None

    def create_instance_profile(self, instance_profile_name: str) -> str | None:
        instance_profile = self._run_json_command(
            f"iam create-instance-profile --instance-profile-name {instance_profile_name}"
        )

        if not instance_profile:
            return None

        return instance_profile["InstanceProfile"]["Arn"]

    def delete_instance_profile(self, instance_profile_name: str, role_name: str):
        self._run_json_command(
            f"iam remove-role-from-instance-profile --instance-profile-name {instance_profile_name}"
            f" --role-name {role_name}"
        )
        self._run_json_command(f"iam delete-instance-profile --instance-profile-name {instance_profile_name}")
        self._run_json_command(f"iam delete-role --role-name {role_name}")

    def add_role_to_instance_profile(self, instance_profile_name: str, role_name: str):
        # there can only be one role associated with iam instance profile
        self._run_command(
            f"iam add-role-to-instance-profile --instance-profile-name {instance_profile_name} --role-name {role_name}"
        )

    def role_exists(self, role_name: str) -> bool:
        """
        Check if the given role exists in the AWS account.
        """
        result = self._run_json_command("iam list-roles")
        roles = result.get("Roles", [])
        for role in roles:
            if role["RoleName"] == role_name:
                return True
        return False

    def _run_json_command(self, command: str):
        aws_cmd = shutil.which("aws")
        code, output, error = self._command_runner(f"{aws_cmd} {command} --profile {self._profile} --output json")
        if code != 0:
            logger.error(error)
            return None
        if output == "":
            return {}
        return json.loads(output)

    def _run_command(self, command: str):
        aws_cmd = shutil.which("aws")
        code, _, error = self._command_runner(f"{aws_cmd} {command} --profile {self._profile} --output json")
        if code != 0:
            logger.error(error)
            return False
        return True

    @staticmethod
    def _get_json_for_cli(input_json: dict) -> str:
        return json.dumps(input_json).replace('\n', '').replace(" ", "")



from __future__ import annotations

import os
from collections import OrderedDict
from collections.abc import Generator
from io import StringIO, BytesIO

from databricks.sdk import WorkspaceClient
from databricks.sdk.service.workspace import ObjectInfo
from databricks.labs.blueprint.paths import WorkspacePath


class _CachedIO:

    def __init__(self, content):
        self._content = content
        self._index = 0

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False

    def read(self, *args, **_kwargs):
        count = -1 if len(args) < 1 or args[0] < 1 else args[0]
        if count == -1:
            return self._content
        start = self._index
        end = self._index + count
        if start >= len(self._content):
            return None
        self._index = self._index + count
        return self._content[start:end]

    def __iter__(self):
        if isinstance(self._content, str):
            yield from StringIO(self._content)
            return
        yield from self._as_string_io().__iter__()

    def with_mode(self, mode: str):
        if 'b' in mode:
            return self._as_bytes_io()
        return self._as_string_io()

    def _as_bytes_io(self):
        if isinstance(self._content, bytes):
            return self
        return BytesIO(self._content.encode("utf-8-sig"))

    def _as_string_io(self):
        if isinstance(self._content, str):
            return self
        return StringIO(self._content.decode("utf-8"))


# lru_cache won't let us invalidate cache entries
# so we provide our own custom lru_cache
class _PathLruCache:

    def __init__(self, max_entries: int):
        self._datas: OrderedDict[str, bytes | str] = OrderedDict()
        self._max_entries = max_entries

    def open(self, cached_path: _CachedPath, mode, buffering, encoding, errors, newline):
        path = str(cached_path)
        if path in self._datas:
            self._datas.move_to_end(path)
            return _CachedIO(self._datas[path]).with_mode(mode)
        io_obj = WorkspacePath.open(cached_path, mode, buffering, encoding, errors, newline)
        # can't read twice from an IO so need to cache data rather than the io object
        data = io_obj.read()
        self._datas[path] = data
        result = _CachedIO(data).with_mode(mode)
        if len(self._datas) > self._max_entries:
            self._datas.popitem(last=False)
        return result

    def clear(self):
        self._datas.clear()

    def remove(self, path: str):
        if path in self._datas:
            self._datas.pop(path)


class _CachedPath(WorkspacePath):
    def __init__(self, cache: _PathLruCache, ws: WorkspaceClient, *args: str | bytes | os.PathLike):
        super().__init__(ws, *args)
        self._cache = cache

    def with_object_info(self, object_info: ObjectInfo):
        self._cached_object_info = object_info
        return self

    def with_segments(self, *path_segments: bytes | str | os.PathLike) -> _CachedPath:
        return type(self)(self._cache, self._ws, *path_segments)

    def iterdir(self) -> Generator[_CachedPath, None, None]:
        for object_info in self._ws.workspace.list(self.as_posix()):
            path = object_info.path
            if path is None:
                msg = f"Cannot initialise without object path: {object_info}"
                raise ValueError(msg)
            child = _CachedPath(self._cache, self._ws, path)
            yield child.with_object_info(object_info)

    def open(
        self,
        mode: str = "r",
        buffering: int = -1,
        encoding: str | None = None,
        errors: str | None = None,
        newline: str | None = None,
    ):
        # only cache reads
        if 'r' in mode:
            return self._cache.open(self, mode, buffering, encoding, errors, newline)
        self._cache.remove(str(self))
        return super().open(mode, buffering, encoding, errors, newline)

    def _cached_open(self, mode: str, buffering: int, encoding: str | None, errors: str | None, newline: str | None):
        return super().open(mode, buffering, encoding, errors, newline)

    # _rename calls unlink so no need to override it
    def unlink(self, missing_ok: bool = False) -> None:
        self._cache.remove(str(self))
        return super().unlink(missing_ok)


class WorkspaceCache:

    def __init__(self, ws: WorkspaceClient, max_entries=2048):
        self._ws = ws
        self._cache = _PathLruCache(max_entries)

    def get_path(self, path: str):
        return _CachedPath(self._cache, self._ws, path)



import datetime
import logging
from dataclasses import dataclass, replace
from collections.abc import Iterable, KeysView

from databricks.labs.lsql.backends import SqlBackend
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import NotFound

from databricks.labs.ucx.framework.crawlers import CrawlerBase
from databricks.labs.ucx.framework.utils import escape_sql_identifier
from databricks.labs.ucx.hive_metastore import TablesCrawler

logger = logging.getLogger(__name__)


@dataclass
class MigrationStatus:
    src_schema: str
    src_table: str
    dst_catalog: str | None = None
    dst_schema: str | None = None
    dst_table: str | None = None
    update_ts: str | None = None

    def destination(self):
        return f"{self.dst_catalog}.{self.dst_schema}.{self.dst_table}".lower()

    @classmethod
    def from_json(cls, raw: dict[str, str]) -> "MigrationStatus":
        return cls(
            src_schema=raw['src_schema'],
            src_table=raw['src_table'],
            dst_catalog=raw.get('dst_catalog', None),
            dst_schema=raw.get('dst_schema', None),
            dst_table=raw.get('dst_table', None),
            update_ts=raw.get('update_ts', None),
        )


@dataclass(frozen=True)
class TableView:
    catalog: str
    schema: str
    name: str

    @property
    def key(self):
        return f"{self.catalog}.{self.schema}.{self.name}".lower()


class MigrationIndex:
    def __init__(self, tables: list[MigrationStatus]):
        self._index = {(ms.src_schema, ms.src_table): ms for ms in tables}

    def is_migrated(self, schema: str, table: str) -> bool:
        """Check if a table is migrated."""
        return self.get(schema, table) is not None

    def get(self, schema: str, table: str) -> MigrationStatus | None:
        """Get the migration status for a table. If the table is not migrated, return None."""
        dst = self._index.get((schema.lower(), table.lower()))
        if not dst or not dst.dst_table:
            return None
        return dst

    def snapshot(self) -> KeysView[tuple[str, str]]:
        return self._index.keys()


class MigrationStatusRefresher(CrawlerBase[MigrationStatus]):
    def __init__(self, ws: WorkspaceClient, sbe: SqlBackend, schema, table_crawler: TablesCrawler):
        super().__init__(sbe, "hive_metastore", schema, "migration_status", MigrationStatus)
        self._ws = ws
        self._table_crawler = table_crawler

    def snapshot(self) -> Iterable[MigrationStatus]:
        return self._snapshot(self._try_fetch, self._crawl)

    def index(self) -> MigrationIndex:
        return MigrationIndex(list(self.snapshot()))

    def get_seen_tables(self) -> dict[str, str]:
        seen_tables: dict[str, str] = {}
        for schema in self._iter_schemas():
            try:
                # ws.tables.list returns Iterator[TableInfo], so we need to convert it to a list in order to catch the exception
                tables = list(self._ws.tables.list(catalog_name=schema.catalog_name, schema_name=schema.name))
            except NotFound:
                logger.warning(
                    f"Schema {schema.catalog_name}.{schema.name} no longer exists. Skipping checking its migration status."
                )
                continue
            for table in tables:
                if not table.properties:
                    continue
                if "upgraded_from" not in table.properties:
                    continue
                if not table.full_name:
                    logger.warning(f"The table {table.name} in {schema.name} has no full name")
                    continue
                seen_tables[table.full_name.lower()] = table.properties["upgraded_from"].lower()
        return seen_tables

    def is_migrated(self, schema: str, table: str) -> bool:
        results = self._backend.fetch(
            f"SHOW TBLPROPERTIES {escape_sql_identifier(schema + '.' + table)} ('upgraded_to')"
        )
        for result in results:
            if "does not have property" in result.value:
                continue
            logger.info(f"{schema}.{table} is set as migrated")
            return True
        logger.info(f"{schema}.{table} is set as not migrated")
        return False

    def _crawl(self) -> Iterable[MigrationStatus]:
        all_tables = self._table_crawler.snapshot()
        reverse_seen = {v: k for k, v in self.get_seen_tables().items()}
        timestamp = datetime.datetime.now(datetime.timezone.utc).timestamp()
        for table in all_tables:
            src_schema = table.database.lower()
            src_table = table.name.lower()
            table_migration_status = MigrationStatus(
                src_schema=src_schema,
                src_table=src_table,
                update_ts=str(timestamp),
            )
            if table.key in reverse_seen and self.is_migrated(src_schema, src_table):
                target_table = reverse_seen[table.key]
                if len(target_table.split(".")) == 3:
                    table_migration_status = replace(
                        table_migration_status,
                        dst_catalog=target_table.split(".")[0],
                        dst_schema=target_table.split(".")[1],
                        dst_table=target_table.split(".")[2],
                    )
            yield table_migration_status

    def _try_fetch(self) -> Iterable[MigrationStatus]:
        for row in self._fetch(f"SELECT * FROM {self._catalog}.{self._schema}.{self._table}"):
            yield MigrationStatus(*row)

    def _iter_schemas(self):
        for catalog in self._ws.catalogs.list():
            try:
                yield from self._ws.schemas.list(catalog_name=catalog.name)
            except NotFound:
                logger.warning(f"Catalog {catalog.name} no longer exists. Skipping checking its migration status.")
                continue



import json
import logging
from collections.abc import Callable
from datetime import timedelta
from functools import partial

from databricks.labs.blueprint.limiter import rate_limited
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import (
    Aborted,
    DeadlineExceeded,
    InternalError,
    NotFound,
    PermissionDenied,
    ResourceConflict,
)
from databricks.sdk.retries import retried
from databricks.sdk.service import iam
from databricks.sdk.service.iam import Group, Patch, PatchSchema

from databricks.labs.ucx.workspace_access.base import AclSupport, Permissions, StaticListing
from databricks.labs.ucx.workspace_access.groups import MigrationState

logger = logging.getLogger(__name__)


class ScimSupport(AclSupport):
    def __init__(
        self,
        ws: WorkspaceClient,
        verify_timeout: timedelta | None = timedelta(minutes=1),
        include_object_permissions: list[str] | None = None,
    ):
        self._ws = ws
        self._verify_timeout = verify_timeout
        self._include_object_permissions = include_object_permissions
        self._snapshot = {}
        # TODO: we may need to inject GroupManager here for proper group listing
        for group in self._ws.groups.list(attributes="id,displayName,meta,roles,entitlements"):
            if not group.display_name:
                continue
            self._snapshot[group.display_name] = group

    @staticmethod
    def _is_item_relevant(item: Permissions, migration_state: MigrationState) -> bool:
        return any(g.id_in_workspace == item.object_id for g in migration_state.groups)

    def get_crawler_tasks(self):
        if self._include_object_permissions:
            for item in StaticListing(self._include_object_permissions, self.object_types()):
                yield partial(self._crawler_task, item.object_id, item.object_type)
            return
        for group in self._snapshot.values():
            meta = group.meta
            if not meta:
                continue
            if meta.resource_type != "WorkspaceGroup":
                continue
            if group.roles and len(group.roles) > 0:
                yield partial(self._crawler_task, group, "roles")
            if group.entitlements and len(group.entitlements) > 0:
                yield partial(self._crawler_task, group, "entitlements")

    def object_types(self) -> set[str]:
        return {"roles", "entitlements"}

    def get_apply_task(self, item: Permissions, migration_state: MigrationState):
        if not self._is_item_relevant(item, migration_state):
            return None
        value = [iam.ComplexValue.from_dict(e) for e in json.loads(item.raw)]
        for migrated_group in migration_state.groups:
            if migrated_group.id_in_workspace != item.object_id:
                continue
            target_group = self._snapshot.get(migrated_group.name_in_account)
            if not target_group:
                msg = f"target group for {migrated_group.name_in_workspace} ({migrated_group.id_in_workspace}) not found: {migrated_group.name_in_account}"
                raise ValueError(msg)
            return partial(self._applier_task, target_group.id, value, item.object_type)
        return None

    def load_for_group(self, group_id: str) -> tuple[set[str], set[str]]:
        roles: set[str] = set()
        entitlements: set[str] = set()
        group = self._safe_get_group(group_id)
        if not group:
            return roles, entitlements
        if group.roles:
            for role in group.roles:
                if not role.value:
                    continue
                roles.add(role.value)
        if group.entitlements:
            for entitlement in group.entitlements:
                if not entitlement.value:
                    continue
                entitlements.add(entitlement.value)
        return roles, entitlements

    @staticmethod
    def _crawler_task(group: iam.Group, property_name: str):
        assert group.id is not None
        return Permissions(
            object_id=group.id,
            object_type=property_name,
            raw=json.dumps([e.as_dict() for e in getattr(group, property_name)]),
        )

    @rate_limited(max_requests=255, burst_period_seconds=60)
    def _verify(self, group_id: str, value: list[iam.ComplexValue], property_name: str):
        # in-flight check for the applied permissions
        # the api might be inconsistent, therefore we need to check that the permissions were applied
        group = self._safe_get_group(group_id)
        if group:
            if property_name == "roles" and group.roles:
                if all(elem in group.roles for elem in value):
                    return True
            if property_name == "entitlements" and group.entitlements:
                if all(elem in group.entitlements for elem in value):
                    return True
            msg = (
                f"Couldn't find role for group {group_id}\n"
                f"acl to be applied={[e.as_dict() for e in value]}\n"
                f"acl found in the object={group.as_dict()}\n"
            )
            raise ValueError(msg)
        return False

    def get_verify_task(self, item: Permissions) -> Callable[[], bool]:
        value = [iam.ComplexValue.from_dict(e) for e in json.loads(item.raw)]
        return partial(self._verify, item.object_id, value, item.object_type)

    @rate_limited(max_requests=10, burst_period_seconds=60)
    def _applier_task(self, group_id: str, value: list[iam.ComplexValue], property_name: str):
        retryable_errors = [ResourceConflict, Aborted, DeadlineExceeded, InternalError]

        operations = [iam.Patch(op=iam.PatchOp.ADD, path=property_name, value=[e.as_dict() for e in value])]
        schemas = [iam.PatchSchema.URN_IETF_PARAMS_SCIM_API_MESSAGES_2_0_PATCH_OP]

        patch_retry_on_value_error = retried(
            on=retryable_errors, timeout=self._verify_timeout  # type: ignore[arg-type]
        )
        patch_retried_check = patch_retry_on_value_error(self._safe_patch_group)
        patch_retried_check(group_id, operations, schemas)

        retry_on_value_error = retried(on=[*retryable_errors, ValueError], timeout=self._verify_timeout)
        retried_check = retry_on_value_error(self._verify)
        return retried_check(group_id, value, property_name)

    def _safe_patch_group(
        self, group_id: str, operations: list[Patch] | None = None, schemas: list[PatchSchema] | None = None
    ):
        try:
            return self._ws.groups.patch(group_id, operations=operations, schemas=schemas)
        except PermissionDenied:
            logger.warning(f"permission denied: {group_id}")
            return None
        except NotFound:
            logger.warning(f"removed on backend: {group_id}")
            return None

    def _safe_get_group(self, group_id: str) -> Group | None:
        try:
            return self._ws.groups.get(group_id)
        except PermissionDenied:
            logger.warning(f"permission denied: {group_id}")
            return None
        except NotFound:
            logger.warning(f"removed on backend: {group_id}")
            return None



from typing import cast

from databricks.sdk.service.workspace import Language

from databricks.labs.ucx.hive_metastore.migration_status import MigrationIndex
from databricks.labs.ucx.source_code.base import (
    Fixer,
    Linter,
    SqlSequentialLinter,
    CurrentSessionState,
    PythonSequentialLinter,
    PythonLinter,
    SqlLinter,
)
from databricks.labs.ucx.source_code.linters.dbfs import DbfsUsageSqlLinter, DBFSUsagePyLinter
from databricks.labs.ucx.source_code.linters.imports import DbutilsPyLinter

from databricks.labs.ucx.source_code.linters.pyspark import SparkSqlPyLinter
from databricks.labs.ucx.source_code.linters.spark_connect import SparkConnectPyLinter
from databricks.labs.ucx.source_code.linters.table_creation import DBRv8d0PyLinter
from databricks.labs.ucx.source_code.queries import FromTableSqlLinter


class LinterContext:
    def __init__(self, index: MigrationIndex | None = None, session_state: CurrentSessionState | None = None):
        self._index = index
        session_state = CurrentSessionState() if not session_state else session_state

        python_linters: list[PythonLinter] = []
        python_fixers: list[Fixer] = []

        sql_linters: list[SqlLinter] = []
        sql_fixers: list[Fixer] = []

        if index is not None:
            from_table = FromTableSqlLinter(index, session_state=session_state)
            sql_linters.append(from_table)
            sql_fixers.append(from_table)
            python_linters.append(SparkSqlPyLinter(from_table, index, session_state))
            python_fixers.append(SparkSqlPyLinter(from_table, index, session_state))

        python_linters += [
            DBFSUsagePyLinter(session_state),
            DBRv8d0PyLinter(dbr_version=session_state.dbr_version),
            SparkConnectPyLinter(session_state),
            DbutilsPyLinter(session_state),
        ]
        sql_linters.append(DbfsUsageSqlLinter())

        self._linters: dict[Language, list[SqlLinter] | list[PythonLinter]] = {
            Language.PYTHON: python_linters,
            Language.SQL: sql_linters,
        }
        self._fixers: dict[Language, list[Fixer]] = {
            Language.PYTHON: python_fixers,
            Language.SQL: sql_fixers,
        }

    def is_supported(self, language: Language) -> bool:
        return language in self._linters and language in self._fixers

    def linter(self, language: Language) -> Linter:
        if language not in self._linters:
            raise ValueError(f"Unsupported language: {language}")
        if language is Language.PYTHON:
            return PythonSequentialLinter(cast(list[PythonLinter], self._linters[language]))
        if language is Language.SQL:
            return SqlSequentialLinter(cast(list[SqlLinter], self._linters[language]))
        raise ValueError(f"Unsupported language: {language}")

    def fixer(self, language: Language, diagnostic_code: str) -> Fixer | None:
        if language not in self._fixers:
            return None
        for fixer in self._fixers[language]:
            if fixer.name() == diagnostic_code:
                return fixer
        return None

    def apply_fixes(self, language: Language, code: str) -> str:
        linter = self.linter(language)
        for advice in linter.lint(code):
            fixer = self.fixer(language, advice.code)
            if fixer:
                code = fixer.apply(code)
        return code



from __future__ import annotations

import abc
import logging
from collections.abc import Callable, Iterable
from pathlib import Path
from typing import TypeVar, cast

from astroid import (  # type: ignore
    Attribute,
    Call,
    Const,
    InferenceError,
    Import,
    ImportFrom,
    Name,
    NodeNG,
)

from databricks.labs.ucx.source_code.base import Advice, Advisory, CurrentSessionState, PythonLinter
from databricks.labs.ucx.source_code.linters.python_ast import Tree, NodeBase, TreeVisitor
from databricks.labs.ucx.source_code.linters.python_infer import InferredValue
from databricks.labs.ucx.source_code.path_lookup import PathLookup

logger = logging.getLogger(__name__)

T = TypeVar("T")
ProblemFactory = Callable[[str, str, NodeNG], T]


class ImportSource(NodeBase):

    @classmethod
    def extract_from_tree(cls, tree: Tree, problem_factory: ProblemFactory) -> tuple[list[ImportSource], list[T]]:
        problems: list[T] = []
        sources: list[ImportSource] = []
        try:  # pylint: disable=too-many-try-statements
            nodes = tree.locate(Import, [])
            for source in cls._make_sources_for_import_nodes(nodes):
                sources.append(source)
            nodes = tree.locate(ImportFrom, [])
            for source in cls._make_sources_for_import_from_nodes(nodes):
                sources.append(source)
            nodes = tree.locate(Call, [("import_module", Attribute), ("importlib", Name)])
            nodes.extend(tree.locate(Call, [("__import__", Attribute), ("importlib", Name)]))
            for source in cls._make_sources_for_import_call_nodes(nodes, problem_factory, problems):
                sources.append(source)
            return sources, problems
        except Exception as e:  # pylint: disable=broad-except
            problem = problem_factory('internal-error', f"While checking imports: {e}", tree.node)
            problems.append(problem)
            return [], problems

    @classmethod
    def _make_sources_for_import_nodes(cls, nodes: list[Import]) -> Iterable[ImportSource]:
        for node in nodes:
            for name, _ in node.names:
                if name is not None:
                    yield ImportSource(node, name)

    @classmethod
    def _make_sources_for_import_from_nodes(cls, nodes: list[ImportFrom]) -> Iterable[ImportSource]:
        for node in nodes:
            yield ImportSource(node, node.modname)

    @classmethod
    def _make_sources_for_import_call_nodes(cls, nodes: list[Call], problem_factory: ProblemFactory, problems: list[T]):
        for node in nodes:
            arg = node.args[0]
            if isinstance(arg, Const):
                yield ImportSource(node, arg.value)
                continue
            problem = problem_factory(
                'dependency-not-constant', "Can't check dependency not provided as a constant", node
            )
            problems.append(problem)

    def __init__(self, node: NodeNG, name: str):
        super().__init__(node)
        self.name = name


class NotebookRunCall(NodeBase):

    def __init__(self, node: Call):
        super().__init__(node)

    def get_notebook_paths(self, session_state: CurrentSessionState) -> tuple[bool, list[str]]:
        """we return multiple paths because astroid can infer them in scenarios such as:
        paths = ["p1", "p2"]
        for path in paths:
            dbutils.notebook.run(path)
        """
        arg = DbutilsPyLinter.get_dbutils_notebook_run_path_arg(self.node)
        try:
            all_inferred = InferredValue.infer_from_node(arg, session_state)
            return self._get_notebook_paths(all_inferred)
        except InferenceError:
            logger.debug(f"Can't infer value(s) of {arg.as_string()}")
            return True, []

    @classmethod
    def _get_notebook_paths(cls, all_inferred: Iterable[InferredValue]) -> tuple[bool, list[str]]:
        has_unresolved = False
        paths: list[str] = []
        for inferred in all_inferred:
            if inferred.is_inferred():
                paths.append(inferred.as_string().strip("'").strip('"'))
                continue
            typenames = [type(node).__name__ for node in inferred.nodes]
            logger.debug(f"Can't compute nodes [{','.join(typenames)}]")
            has_unresolved = True
        return has_unresolved, paths


class DbutilsPyLinter(PythonLinter):

    def __init__(self, session_state: CurrentSessionState):
        self._session_state = session_state

    def lint_tree(self, tree: Tree) -> Iterable[Advice]:
        nodes = self.list_dbutils_notebook_run_calls(tree)
        for node in nodes:
            yield from self._raise_advice_if_unresolved(node.node, self._session_state)

    @classmethod
    def _raise_advice_if_unresolved(cls, node: NodeNG, session_state: CurrentSessionState) -> Iterable[Advice]:
        assert isinstance(node, Call)
        call = NotebookRunCall(cast(Call, node))
        has_unresolved, _ = call.get_notebook_paths(session_state)
        if has_unresolved:
            yield Advisory.from_node(
                code='notebook-run-cannot-compute-value',
                message="Path for 'dbutils.notebook.run' cannot be computed and requires adjusting the notebook path(s)",
                node=node,
            )

    @staticmethod
    def get_dbutils_notebook_run_path_arg(node: Call):
        if len(node.args) > 0:
            return node.args[0]
        arg = next(kw for kw in node.keywords if kw.arg == "path")
        return arg.value if arg is not None else None

    @staticmethod
    def list_dbutils_notebook_run_calls(tree: Tree) -> list[NotebookRunCall]:
        calls = tree.locate(Call, [("run", Attribute), ("notebook", Attribute), ("dbutils", Name)])
        return [NotebookRunCall(call) for call in calls]


class SysPathChange(NodeBase, abc.ABC):

    @staticmethod
    def extract_from_tree(session_state: CurrentSessionState, tree: Tree) -> list[SysPathChange]:
        visitor = SysPathChangesVisitor(session_state)
        visitor.visit(tree.node)
        return visitor.sys_path_changes

    def __init__(self, node: NodeNG, path: str, is_append: bool):
        super().__init__(node)
        self._path = path
        self._is_append = is_append

    @property
    def path(self):
        return self._path

    def apply_to(self, path_lookup: PathLookup):
        path = Path(self._path)
        if not path.is_absolute():
            path = path_lookup.cwd / path
        if self._is_append:
            path_lookup.append_path(path)
            return
        path_lookup.prepend_path(path)


class AbsolutePath(SysPathChange):
    # path directly added to sys.path
    pass


class RelativePath(SysPathChange):
    # path added to sys.path using os.path.abspath
    pass


class UnresolvedPath(SysPathChange):
    # path added to sys.path that cannot be inferred
    pass


class SysPathChangesVisitor(TreeVisitor):

    def __init__(self, session_state: CurrentSessionState) -> None:
        super()
        self._session_state = session_state
        self._aliases: dict[str, str] = {}
        self.sys_path_changes: list[SysPathChange] = []

    def visit_import(self, node: Import):
        for name, alias in node.names:
            if alias is None or name not in {"sys", "os"}:
                continue
            self._aliases[name] = alias

    def visit_importfrom(self, node: ImportFrom):
        interesting_aliases = [("sys", "path"), ("os", "path"), ("os.path", "abspath")]
        interesting_alias = next((t for t in interesting_aliases if t[0] == node.modname), None)
        if interesting_alias is None:
            return
        for name, alias in node.names:
            if name == interesting_alias[1]:
                self._aliases[f"{node.modname}.{interesting_alias[1]}"] = alias or name
                break

    def visit_call(self, node: Call):
        func = cast(Attribute, node.func)
        # check for 'sys.path.append'
        if not (
            self._match_aliases(func, ["sys", "path", "append"]) or self._match_aliases(func, ["sys", "path", "insert"])
        ):
            return
        is_append = func.attrname == "append"
        changed = node.args[0] if is_append else node.args[1]
        relative = False
        if isinstance(changed, Call):
            if not self._match_aliases(changed.func, ["os", "path", "abspath"]):
                return
            relative = True
            changed = changed.args[0]
        try:
            for inferred in InferredValue.infer_from_node(changed, self._session_state):
                self._visit_inferred(changed, inferred, relative, is_append)
        except InferenceError:
            self.sys_path_changes.append(UnresolvedPath(changed, changed.as_string(), is_append))

    def _visit_inferred(self, changed: NodeNG, inferred: InferredValue, is_relative: bool, is_append: bool):
        if not inferred.is_inferred():
            self.sys_path_changes.append(UnresolvedPath(changed, changed.as_string(), is_append))
            return
        if is_relative:
            self.sys_path_changes.append(RelativePath(changed, inferred.as_string(), is_append))
        else:
            self.sys_path_changes.append(AbsolutePath(changed, inferred.as_string(), is_append))

    def _match_aliases(self, node: NodeNG, names: list[str]):
        if isinstance(node, Attribute):
            if node.attrname != names[-1]:
                return False
            if len(names) == 1:
                return True
            return self._match_aliases(node.expr, names[0 : len(names) - 1])
        if isinstance(node, Name):
            full_name = ".".join(names)
            alias = self._aliases.get(full_name, full_name)
            return node.name == alias
        return False



import logging
from collections.abc import Callable, Iterable, Sequence
from typing import ClassVar, Generic, Protocol, TypeVar

from databricks.labs.lsql.backends import SqlBackend
from databricks.sdk.errors import NotFound

from databricks.labs.ucx.framework.utils import escape_sql_identifier

logger = logging.getLogger(__name__)


class DataclassInstance(Protocol):
    __dataclass_fields__: ClassVar[dict]


Result = TypeVar("Result", bound=DataclassInstance)
Dataclass = type[DataclassInstance]
ResultFn = Callable[[], Iterable[Result]]


class CrawlerBase(Generic[Result]):
    def __init__(self, backend: SqlBackend, catalog: str, schema: str, table: str, klass: type[Result]):
        """
        Initializes a CrawlerBase instance.

        Args:
            backend (SqlBackend): The backend that executes SQL queries:
                Statement Execution API or Databricks Runtime.
            catalog (str): The catalog name for the inventory persistence.
            schema: The schema name for the inventory persistence.
            table: The table name for the inventory persistence.
        """
        self._catalog = self._valid(catalog)
        self._schema = self._valid(schema)
        self._table = self._valid(table)
        self._backend = backend
        self._fetch = backend.fetch
        self._exec = backend.execute
        self._klass = klass

    @property
    def full_name(self) -> str:
        """
        Generates the full name of the table.

        Returns:
            str: The full table name.
        """
        return f"{self._catalog}.{self._schema}.{self._table}"

    def reset(self):
        """
        Delete the content of the inventory table.
        The next call to `snapshot` will re-populate the table.
        """
        self._exec(f"TRUNCATE TABLE {escape_sql_identifier(self.full_name)}")

    @staticmethod
    def _valid(name: str) -> str:
        """
        Validates that the provided name does not contain dots.

        Args:
            name (str): The name to be validated.

        Returns:
            str: The validated name.

        Raises:
            ValueError: If the name contains dots.
        """
        if "." in name:
            msg = f"no dots allowed in `{name}`"
            raise ValueError(msg)
        return name

    @classmethod
    def _try_valid(cls, name: str | None):
        """
        Tries to validate a name. If None, returns None.

        Args:
            name (str): The name to be validated.

        Returns:
            str or None: The validated name or None.
        """
        if name is None:
            return None
        return cls._valid(name)

    def _snapshot(self, fetcher: ResultFn, loader: ResultFn) -> list[Result]:
        """
        Tries to load dataset of records with `fetcher` function, otherwise automatically creates
        a table with the schema defined in the class of the first row and executes `loader` function
        to populate the dataset.

        Args:
            fetcher: A function to fetch existing data.
            loader: A function to load new data.

        Exceptions:
        - If a runtime error occurs during fetching (other than "TABLE_OR_VIEW_NOT_FOUND"), the original error is
          re-raised.

        Returns:
        list[any]: A list of data records, either fetched or loaded.
        """
        logger.debug(f"[{self.full_name}] fetching {self._table} inventory")
        try:
            cached_results = list(fetcher())
            if len(cached_results) > 0:
                return cached_results
        except NotFound:
            pass
        logger.debug(f"[{self.full_name}] crawling new batch for {self._table}")
        loaded_records = list(loader())
        self._append_records(loaded_records)
        return loaded_records

    def _append_records(self, items: Sequence[Result]):
        logger.debug(f"[{self.full_name}] found {len(items)} new records for {self._table}")
        self._backend.save_table(self.full_name, items, self._klass, mode="append")



from .base import (
    SchemaComparator,
    SchemaComparisonEntry,
    SchemaComparisonResult,
    TableMetadataRetriever,
    ColumnMetadata,
    TableIdentifier,
)


class StandardSchemaComparator(SchemaComparator):
    def __init__(self, metadata_retriever: TableMetadataRetriever):
        self._metadata_retriever = metadata_retriever

    def compare_schema(self, source: TableIdentifier, target: TableIdentifier) -> SchemaComparisonResult:
        """
        This method compares the schema of two tables. It takes two TableIdentifier objects as input, which represent
        the source and target tables for which the schemas are to be compared.

        Note: This method does not handle exceptions raised during the execution of the SQL query or the retrieval
        of the table metadata. These exceptions are expected to be handled by the caller in a manner appropriate for
        their context.
        """
        comparison_result = self._eval_schema_diffs(source, target)
        is_matching = all(entry.is_matching for entry in comparison_result)
        return SchemaComparisonResult(is_matching, comparison_result)

    def _eval_schema_diffs(self, source: TableIdentifier, target: TableIdentifier) -> list[SchemaComparisonEntry]:
        source_metadata = self._metadata_retriever.get_metadata(source)
        target_metadata = self._metadata_retriever.get_metadata(target)
        # Combine the sets of column names for both the source and target tables
        # to create a set of all unique column names from both tables.
        source_column_names = {column.name for column in source_metadata.columns}
        target_column_names = {column.name for column in target_metadata.columns}
        all_column_names = source_column_names.union(target_column_names)
        comparison_result = []
        # Compare the column metadata from each table with a logic similar to full outer join.
        for field_name in sorted(all_column_names):
            source_col = source_metadata.get_column_metadata(field_name)
            target_col = target_metadata.get_column_metadata(field_name)
            comparison_result.append(_build_comparison_result_entry(source_col, target_col))
        return comparison_result


def _build_comparison_result_entry(
    source_col: ColumnMetadata | None,
    target_col: ColumnMetadata | None,
) -> SchemaComparisonEntry:
    if source_col and target_col:
        is_matching = source_col == target_col
        notes = None
    else:
        is_matching = False
        notes = "Column is missing in " + ("target" if source_col else "source")

    return SchemaComparisonEntry(
        source_column=source_col.name if source_col else None,
        source_datatype=source_col.data_type if source_col else None,
        target_column=target_col.name if target_col else None,
        target_datatype=target_col.data_type if target_col else None,
        is_matching=is_matching,
        notes=notes,
    )



import collections
import dataclasses
import functools
import json
from collections.abc import Callable, Iterator
from datetime import timedelta
from functools import partial

from databricks.labs.lsql.backends import SqlBackend
from databricks.sdk.errors import NotFound
from databricks.sdk.retries import retried

from databricks.labs.ucx.hive_metastore.grants import Grant, GrantsCrawler
from databricks.labs.ucx.workspace_access.base import AclSupport, Permissions, StaticListing
from databricks.labs.ucx.workspace_access.groups import MigrationState


class TableAclSupport(AclSupport):
    def __init__(
        self,
        grants_crawler: GrantsCrawler,
        sql_backend: SqlBackend,
        verify_timeout: timedelta | None = timedelta(minutes=1),
        # this parameter is for testing scenarios only - [{object_type}:{object_id}]
        # it will use StaticListing class to return only object ids that has the same object type
        include_object_permissions: list[str] | None = None,
    ):
        self._grants_crawler = grants_crawler
        self._sql_backend = sql_backend
        self._verify_timeout = verify_timeout
        self._include_object_permissions = include_object_permissions

    def get_crawler_tasks(self) -> Iterator[Callable[..., Permissions | None]]:
        # Table ACL permissions (grant/revoke and ownership) are not atomic. When granting the permissions,
        # the service would first get all existing permissions, append with the new permissions,
        # and set the full list in the database. If there are concurrent grant requests,
        # both requests might succeed and emit the audit logs, but what actually happens could be that
        # the new permission list from one request overrides the other one, causing permissions loss.
        # More info here: https://databricks.atlassian.net/browse/ES-908737
        #
        # Below optimization mitigates the issue by folding all action types (grants and own permissions)
        # for the same principal, object_id and object_type into one grant with comma separated list of action types.
        #
        # For example, the following table grants:
        # * GRANT SELECT ON TABLE hive_metastore.db_a.table_a TO group_a
        # * GRANT MODIFY ON TABLE hive_metastore.db_a.table_a TO group_a
        # will be folded and executed in one statement/transaction:
        # * GRANT SELECT, MODIFY ON TABLE hive_metastore.db_a.table_a TO group_a
        # The exception is the "OWN" permission which must be set using a separate ALTER statement.

        include_objects = set[tuple[str, str]]()
        if self._include_object_permissions:
            for item in StaticListing(self._include_object_permissions, self.object_types()):
                include_objects.add((item.object_type, item.object_id))

        folded_actions = collections.defaultdict(set)
        for grant in self._grants_crawler.snapshot():
            type_and_key = grant.this_type_and_key()
            if include_objects and type_and_key not in include_objects:
                continue
            key = (grant.principal, type_and_key)
            folded_actions[key].add(grant.action_type)

        def inner(object_type: str, object_id: str, grant: Grant) -> Permissions:
            return Permissions(object_type=object_type, object_id=object_id, raw=json.dumps(dataclasses.asdict(grant)))

        for (principal, (object_type, object_id)), actions in folded_actions.items():
            grant = self._from_reduced(object_type, object_id, principal, ", ".join(sorted(actions)))
            yield functools.partial(inner, object_type=object_type, object_id=object_id, grant=grant)

    def _from_reduced(self, object_type: str, object_id: str, principal: str, action_type: str):
        match object_type:
            case "TABLE":
                catalog, database, table = object_id.split(".")
                return Grant(
                    principal=principal, action_type=action_type, catalog=catalog, database=database, table=table
                )
            case "VIEW":
                catalog, database, view = object_id.split(".")
                return Grant(
                    principal=principal, action_type=action_type, catalog=catalog, database=database, view=view
                )
            case "DATABASE":
                catalog, database = object_id.split(".")
                return Grant(principal=principal, action_type=action_type, catalog=catalog, database=database)
            case "CATALOG":
                catalog = object_id
                return Grant(principal=principal, action_type=action_type, catalog=catalog)
            case "FUNCTION":
                catalog, database, udf = object_id.split(".")
                return Grant(principal=principal, action_type=action_type, catalog=catalog, database=database, udf=udf)
            case "ANONYMOUS FUNCTION":
                catalog = object_id
                return Grant(principal=principal, action_type=action_type, catalog=catalog, anonymous_function=True)
            case "ANY FILE":
                catalog = object_id
                return Grant(principal=principal, action_type=action_type, catalog=catalog, any_file=True)

    def object_types(self) -> set[str]:
        return {"TABLE", "DATABASE", "VIEW", "CATALOG", "FUNCTION", "ANONYMOUS FUNCTION", "ANY FILE"}

    def get_apply_task(self, item: Permissions, migration_state: MigrationState):
        grant = Grant(**json.loads(item.raw))
        target_principal = migration_state.get_target_principal(grant.principal)
        if target_principal is None:
            # this is a grant for user, service principal, or irrelevant group
            return None
        target_grant = dataclasses.replace(grant, principal=target_principal)
        # this has to be executed on tacl cluster, otherwise - use SQLExecutionAPI backend & Warehouse
        return partial(self._apply_grant, target_grant)

    def _apply_grant(self, grant: Grant):
        """
        Apply grants and OWN permission serially for the same principal, object_id and object_type.
        Executing grants (using GRANT statement) and OWN permission (using ALTER statement) concurrently
        could cause permission lost due to limitations in the Table ACLs.
        More info here: https://databricks.atlassian.net/browse/ES-976290
        """
        for sql in grant.hive_grant_sql():
            self._sql_backend.execute(sql)

        object_type, object_id = grant.this_type_and_key()
        retry_on_value_error = retried(on=[NotFound], timeout=self._verify_timeout)
        retried_check = retry_on_value_error(self._verify)
        return retried_check(object_type, object_id, grant)

    def _verify(self, object_type: str, object_id: str, acl: Grant) -> bool:
        grant_dict = dataclasses.asdict(acl)
        del grant_dict["action_type"]
        del grant_dict["principal"]
        grants_on_object = self._grants_crawler.grants(**grant_dict)

        if grants_on_object:
            on_current_principal = [grant.action_type for grant in grants_on_object if grant.principal == acl.principal]
            acl_action_types = acl.action_type.split(", ")
            if all(action_type in on_current_principal for action_type in acl_action_types):
                return True
            msg = (
                f"Couldn't find permission for object type {object_type}, id {object_id} and principal {acl.principal}\n"
                f"acl to be applied={acl_action_types}\n"
                f"acl found in the object={on_current_principal}\n"
            )
            raise NotFound(msg)
        return False

    def get_verify_task(self, item: Permissions) -> Callable[[], bool]:
        grant = Grant(**json.loads(item.raw))
        return partial(self._verify, item.object_type, item.object_id, grant)
